#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
This module contains the interface to access Rubrix's REST API.
"""


import logging
import os
import re
from typing import Iterable

import pandas
import pkg_resources

from rubrix._constants import DEFAULT_API_KEY
from rubrix.client import RubrixClient
from rubrix.client.models import *
from rubrix.monitoring.model_monitor import monitor

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    # package is not installed
    pass

_LOGGER = logging.getLogger(__name__)

_client: Optional[
    RubrixClient
] = None  # Client will be stored here to pass it through functions


def _client_instance() -> RubrixClient:
    """Checks module instance client and init if not initialized."""

    global _client
    # Calling a by-default-init if it was not called before
    if _client is None:
        init()
    return _client


def init(
    api_url: Optional[str] = None,
    api_key: Optional[str] = None,
    workspace: Optional[str] = None,
    timeout: int = 60,
) -> None:
    """Init the python client.

    Passing an api_url disables environment variable reading, which will provide
    default values.

    Args:
        api_url: Address of the REST API. If `None` (default) and the env variable ``RUBRIX_API_URL`` is not set,
            it will default to `http://localhost:6900`.
        api_key: Authentification key for the REST API. If `None` (default) and the env variable ``RUBRIX_API_KEY``
            is not set, it will default to `rubrix.apikey`.
        workspace: The workspace to which records will be logged/loaded. If `None` (default) and the 
            env variable ``RUBRIX_WORKSPACE`` is not set, it will default to the private user workspace.
        timeout: Wait `timeout` seconds for the connection to timeout. Default: 60.

    Examples:
        >>> import rubrix as rb
        >>> rb.init(api_url="http://localhost:9090", api_key="4AkeAPIk3Y")
    """

    global _client

    final_api_url = api_url or os.getenv("RUBRIX_API_URL", "http://localhost:6900")

    # Checking that the api_url does not end in '/'
    final_api_url = re.sub(r"\/$", "", final_api_url)

    # If an api_url is passed, tokens obtained via environ vars are disabled
    final_key = api_key or os.getenv("RUBRIX_API_KEY", DEFAULT_API_KEY)

    workspace = workspace or os.getenv("RUBRIX_WORKSPACE")

    _LOGGER.info(f"Rubrix has been initialized on {final_api_url}")

    _client = RubrixClient(
        api_url=final_api_url,
        api_key=final_key,
        workspace=workspace,
        timeout=timeout,
    )


def get_workspace() -> str:
    """Returns the name of the active workspace for the current client session.

    Returns:
        The name of the active workspace as a string.
    """
    return _client_instance().active_workspace


def set_workspace(ws: str) -> None:
    """Sets the active workspace for the current client session.

    Args:
        ws: The new workspace
    """
    _client_instance().set_workspace(ws)


def log(
    records: Union[Record, Iterable[Record]],
    name: str,
    tags: Optional[Dict[str, str]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    chunk_size: int = 500,
    verbose: bool = True,
) -> BulkResponse:
    """Log Records to Rubrix.

    Args:
        records: The record or an iterable of records.
        name: The dataset name.
        tags: A dictionary of tags related to the dataset.
        metadata: A dictionary of extra info for the dataset.
        chunk_size: The chunk size for a data bulk.
        verbose: If True, shows a progress bar and prints out a quick summary at the end.

    Returns:
        Summary of the response from the REST API

    Examples:
        >>> import rubrix as rb
        >>> record = rb.TextClassificationRecord(
        ...     inputs={"text": "my first rubrix example"},
        ...     prediction=[('spam', 0.8), ('ham', 0.2)]
        ... )
        >>> response = rb.log(record, name="example-dataset")
    """
    # noinspection PyTypeChecker,PydanticTypeChecker
    return _client_instance().log(
        records=records,
        name=name,
        tags=tags,
        metadata=metadata,
        chunk_size=chunk_size,
        verbose=verbose,
    )


def copy(dataset: str, name_of_copy: str, workspace: str = None):
    """Creates a copy of a dataset including its tags and metadata

    Args:
        dataset: Name of the source dataset
        name_of_copy: Name of the copied dataset
        workspace: If provided, dataset will be copied to that workspace

    Examples:
        >>> import rubrix as rb
        >>> rb.copy("my_dataset", name_of_copy="new_dataset")
        >>> dataframe = rb.load("new_dataset")
    """
    _client_instance().copy(source=dataset, target=name_of_copy, target_workspace=workspace)


def load(
    name: str,
    query: Optional[str] = None,
    ids: Optional[List[Union[str, int]]] = None,
    limit: Optional[int] = None,
    as_pandas: bool = True,
) -> Union[pandas.DataFrame, List[Record]]:
    """Loads a dataset as a pandas DataFrame or a list of records.

    Args:
        name: The dataset name.
        query: An ElasticSearch query with the
            `query string syntax <https://rubrix.readthedocs.io/en/stable/reference/rubrix_webapp_reference.html#search-input>`_
        ids: If provided, load dataset records with given ids.
        limit: The number of records to retrieve.
        as_pandas: If True, return a pandas DataFrame. If False, return a list of records.

    Returns:
        The dataset as a pandas Dataframe or a list of records.

    Examples:
        >>> import rubrix as rb
        >>> dataframe = rb.load(name="example-dataset")
    """
    return _client_instance().load(
        name=name, query=query, limit=limit, ids=ids, as_pandas=as_pandas
    )


def delete(name: str) -> None:
    """Delete a dataset.

    Args:
        name: The dataset name.

    Examples:
        >>> import rubrix as rb
        >>> rb.delete(name="example-dataset")
    """
    _client_instance().delete(name=name)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json

import numpy
import pytest
from pydantic import ValidationError
from rubrix._constants import MAX_KEYWORD_LENGTH

from rubrix.client.models import Text2TextRecord, TextClassificationRecord
from rubrix.client.models import TokenClassificationRecord


@pytest.mark.parametrize(
    ("annotation", "status", "expected_status"),
    [
        (None, None, "Default"),
        ("test", None, "Validated"),
        (None, "Discarded", "Discarded"),
        ("test", "Discarded", "Discarded"),
    ],
)
def test_text_classification_record(annotation, status, expected_status):
    """Just testing its dynamic defaults"""
    record = TextClassificationRecord(
        inputs={"text": "test"}, annotation=annotation, status=status
    )
    assert record.status == expected_status


def test_text_classification_input_string():
    assert TextClassificationRecord(inputs="A text") == TextClassificationRecord(
        inputs=dict(text="A text")
    )

    assert TextClassificationRecord(
        inputs=["A text", "another text"]
    ) == TextClassificationRecord(inputs=dict(text=["A text", "another text"]))


@pytest.mark.parametrize(
    ("annotation", "status", "expected_status"),
    [
        (None, None, "Default"),
        ([("test", 0, 5)], None, "Validated"),
        (None, "Discarded", "Discarded"),
        ([("test", 0, 5)], "Discarded", "Discarded"),
    ],
)
def test_token_classification_record(annotation, status, expected_status):
    """Just testing its dynamic defaults"""
    record = TokenClassificationRecord(
        text="test text", tokens=["test", "text"], annotation=annotation, status=status
    )
    assert record.status == expected_status


def test_text_classification_record_none_inputs():
    """Test validation error for None in inputs"""
    with pytest.raises(ValidationError):
        TextClassificationRecord(inputs={"text": None})


def test_metadata_values_length():
    text = "oh yeah!"
    metadata = {"too_long": "a" * 200}

    record = TextClassificationRecord(inputs={"text": text}, metadata=metadata)
    assert len(record.metadata["too_long"]) == MAX_KEYWORD_LENGTH

    record = TokenClassificationRecord(
        text=text, tokens=text.split(), metadata=metadata
    )
    assert len(record.metadata["too_long"]) == MAX_KEYWORD_LENGTH


def test_model_serialization_with_numpy_nan():
    record = Text2TextRecord(
        text="My name is Sarah and I love my dog.", metadata={"nan": numpy.nan}
    )

    json_record = json.loads(record.json())



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
import datetime
from typing import Any, Dict, Iterable, List, Optional, Type, TypeVar

from fastapi import Depends

from rubrix.server.commons.es_helpers import (
    DATASETS_RECORDS_INDEX_TEMPLATE,
    aggregations,
    parse_aggregations,
)
from rubrix.server.commons.es_wrapper import ElasticsearchWrapper, create_es_wrapper
from rubrix.server.commons.helpers import unflatten_dict
from rubrix.server.commons.settings import settings
from rubrix.server.datasets.dao import (
    DATASETS_RECORDS_INDEX_NAME,
    dataset_records_index,
)
from rubrix.server.datasets.model import BaseDatasetDB
from rubrix.server.tasks.commons import BaseRecord
from rubrix.server.tasks.commons.dao.model import RecordSearch, RecordSearchResults

DBRecord = TypeVar("DBRecord", bound=BaseRecord)


@dataclasses.dataclass
class _IndexTemplateExtensions:

    analyzers: List[Dict[str, Any]] = dataclasses.field(default_factory=list)
    properties: List[Dict[str, Any]] = dataclasses.field(default_factory=list)
    dynamic_templates: List[Dict[str, Any]] = dataclasses.field(default_factory=list)


_extensions = _IndexTemplateExtensions()


def extends_index_properties(extended_properties: Dict[str, Any]):
    """
    Add explict properties configuration to rubrix index template

    See https://www.elastic.co/guide/en/elasticsearch/reference/current/explicit-mapping.html

    Parameters
    ----------
    extended_properties:
        The properties dictionary configuration. Several properties could be configured here

    """
    _extensions.properties.append(extended_properties)


def extends_index_dynamic_templates(*templates: Dict[str, Any]):
    """
    Add dynamic mapping template configuration to rubrix index template

    See https://www.elastic.co/guide/en/elasticsearch/reference/7.x/dynamic-templates.html#dynamic-templates

    Parameters
    ----------
    templates:
        One or several mapping templates
    """
    _extensions.dynamic_templates.extend(templates)


def extends_index_analyzers(analyzers: Dict[str, Any]):
    """
    Add index analysis configuration to rubrix index template

    See https://www.elastic.co/guide/en/elasticsearch/reference/current/analyzer.html

    Parameters
    ----------
    analyzers:
        The analyzers configuration. Several analyzers could be configured here
    """
    _extensions.analyzers.append(analyzers)


class DatasetRecordsDAO:
    """Datasets records DAO"""

    def __init__(self, es: ElasticsearchWrapper):
        self._es = es
        self.init()

    def init(self):
        """Initializes dataset records dao. Used on app startup"""

        template = DATASETS_RECORDS_INDEX_TEMPLATE.copy()

        if _extensions.analyzers:
            for analyzer in _extensions.analyzers:
                template["settings"]["analysis"]["analyzer"].update(analyzer)

        if _extensions.dynamic_templates:
            for dynamic_template in _extensions.dynamic_templates:
                template["mappings"]["dynamic_templates"].append(dynamic_template)

        if _extensions.properties:
            for property in _extensions.properties:
                template["mappings"]["properties"].update(property)

        self._es.create_index_template(
            name=DATASETS_RECORDS_INDEX_NAME,
            template=template,
            force_recreate=not settings.disable_es_index_template_creation,
        )

    def add_records(
        self,
        dataset: BaseDatasetDB,
        records: List[BaseRecord],
        record_class: Type[DBRecord],
    ) -> int:
        """
        Add records to dataset

        Parameters
        ----------
        dataset:
            The dataset
        records:
            The list of records
        record_class:
            Record class used to convert records to
        Returns
        -------
            The number of failed records

        """

        now = None
        documents = []
        metadata_values = {}

        if "last_updated" in record_class.schema()["properties"]:
            now = datetime.datetime.utcnow()

        for r in records:
            metadata_values.update(r.metadata or {})
            db_record = record_class.parse_obj(r)
            if now:
                db_record.last_updated = now
            documents.append(db_record.dict(exclude_none=False))

        index_name = dataset_records_index(dataset.id)

        self._es.create_index(index=index_name)
        self._configure_metadata_fields(index_name, metadata_values)
        return self._es.add_documents(
            index=index_name,
            documents=documents,
            doc_id=lambda _record: _record.get("id"),
        )

    def search_records(
        self,
        dataset: BaseDatasetDB,
        search: Optional[RecordSearch] = None,
        size: int = 100,
        record_from: int = 0,
    ) -> RecordSearchResults:
        """
        SearchRequest records under a dataset given a search parameters.

        Parameters
        ----------
        dataset:
            The dataset
        search:
            The search params
        size:
            Number of records to retrieve (for pagination)
        record_from:
            Record from witch retrieve records (for pagination)
        Returns
        -------
            The search result

        """
        search = search or RecordSearch()
        records_index = dataset_records_index(dataset.id)

        aggregation_requests = (
            {**(search.aggregations or {})} if record_from == 0 else None
        )

        if aggregation_requests is not None and search.include_default_aggregations:
            aggregation_requests.update(
                {
                    **aggregations.predicted_as(),
                    **aggregations.predicted_by(),
                    **aggregations.annotated_as(),
                    **aggregations.annotated_by(),
                    **aggregations.status(),
                    **aggregations.predicted(),
                    **aggregations.words_cloud(),
                    **aggregations.score(),
                    **aggregations.custom_fields(
                        self._es.get_field_mapping(
                            index=records_index, field_name="metadata.*"
                        )
                    ),
                }
            )

        if record_from > 0:
            aggregation_requests = None

        es_query = {
            "size": size,
            "from": record_from,
            "query": search.query or {"match_all": {}},
            "sort": search.sort or [{"_id": {"order": "asc"}}],
            "aggs": aggregation_requests or {},
        }
        results = self._es.search(index=records_index, query=es_query, size=size)

        hits = results["hits"]
        total = hits["total"]
        docs = hits["hits"]
        search_aggregations = results.get("aggregations", {})

        result = RecordSearchResults(
            total=total,
            records=list(map(self.esdoc2record, docs)),
        )
        if search_aggregations:
            parsed_aggregations = parse_aggregations(search_aggregations)
            parsed_aggregations = unflatten_dict(
                parsed_aggregations, stop_keys=["metadata"]
            )

            result.words = parsed_aggregations.pop("words", {})
            result.metadata = parsed_aggregations.pop("metadata", {})
            result.aggregations = parsed_aggregations

        return result

    def scan_dataset(
        self,
        dataset: BaseDatasetDB,
        search: Optional[RecordSearch] = None,
    ) -> Iterable[Dict[str, Any]]:
        """
        Iterates over a dataset records

        Parameters
        ----------
        dataset:
            The dataset
        search:
            The search parameters. Optional

        Returns
        -------
            An iterable over found dataset records
        """
        search = search or RecordSearch()
        es_query = {
            "query": search.query,
        }
        docs = self._es.list_documents(
            dataset_records_index(dataset.id), query=es_query
        )
        for doc in docs:
            yield self.esdoc2record(doc)

    def esdoc2record(self, doc):
        return {**doc["_source"], "id": doc["_id"]}

    def _configure_metadata_fields(self, index: str, metadata_values: Dict[str, Any]):
        def detect_nested_type(v: Any) -> bool:
            """Returns True if value match as nested value"""
            return isinstance(v, list) and isinstance(v[0], dict)

        for field, value in metadata_values.items():
            if detect_nested_type(value):
                self._es.create_field_mapping(
                    index,
                    field_name=f"metadata.{field}",
                    type="nested",
                    include_in_root=True,
                )


_instance: Optional[DatasetRecordsDAO] = None


def dataset_records_dao(
    es: ElasticsearchWrapper = Depends(create_es_wrapper),
) -> DatasetRecordsDAO:
    """
    Creates a dataset records dao instance

    Parameters
    ----------
    es:
        The elasticserach wrapper dependency

    """
    global _instance
    if not _instance:
        _instance = DatasetRecordsDAO(es)
    return _instance



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional, Type

from fastapi import HTTPException, Request, status
from fastapi.encoders import jsonable_encoder
from fastapi.exception_handlers import http_exception_handler
from pydantic import BaseModel, ValidationError


class ErrorMessage(BaseModel):
    """
    Generic error class. This class is needed only for openapi documentation
    purposes.

    If some field in additional http errors changes, these changes
    should be reflected in the error model.

    Attributes:
    -----------

    detail:
        The error message

    """

    detail: str


class UnauthorizedError(HTTPException):
    """Unauthorized error"""

    def __init__(self):
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials",
            headers={"WWW-Authenticate": "Bearer"},
        )


class InactiveUserError(HTTPException):
    """Inactive user error"""

    def __init__(self):
        super().__init__(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Inactive user"
        )


class ForbiddenOperationError(HTTPException):
    """Forbidden operation"""

    def __init__(self, message: Optional[str] = None):
        super().__init__(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=message or "Operation not allowed",
        )


class WrongInputParamError(HTTPException):
    """Error related with input params in general"""

    def __init__(self, detail: str):
        super().__init__(status_code=status.HTTP_400_BAD_REQUEST, detail=detail)


class InvalidTextSearchError(HTTPException):
    """Error related with input params in search"""

    def __init__(self, detail: str):
        super().__init__(status_code=status.HTTP_400_BAD_REQUEST, detail=detail)


class EntityAlreadyExistsError(HTTPException):
    """Error raised when entity was created"""

    def __init__(self, name: str, type: Type, workspace: Optional[str] = None):
        self.name = name
        self.type = Type
        self.workspace = workspace

        msg = f"Already created entity {name} of type {type.__name__}"
        if self.workspace:
            msg += f" in {workspace} workspace."

        super().__init__(
            status_code=status.HTTP_409_CONFLICT,
            detail=msg,
        )


class WrongTaskError(HTTPException):
    """Error raised when provided task cannot be processed with requested entity"""

    def __init__(self, detail: str):
        super().__init__(detail=detail, status_code=status.HTTP_400_BAD_REQUEST)


class EntityNotFoundError(HTTPException):
    """Error raised when entity not found"""

    def __init__(self, name: str, type: Type):
        self.name = name
        self.type = Type

        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Not found entity {name} of type {type.__name__}",
        )


class GenericValidationError(HTTPException):
    """Generic data validation error out of request"""

    def __init__(self, error: ValidationError):
        super().__init__(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(error.errors()),
        )


class GenericError(HTTPException):
    """Generic error"""

    def __init__(self, exception: Exception):
        super().__init__(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(exception)
        )


async def common_exception_handler(request: Request, error: Exception):
    """Wraps errors as custom generic error"""
    return await http_exception_handler(request, GenericError(error))


async def validation_exception_handler(request: Request, error: ValidationError):
    """Wraps pydantic errors"""
    return await http_exception_handler(request, GenericValidationError(error))



from itertools import zip_longest
from typing import Any, Dict, List, Optional

from elasticsearch import Elasticsearch
from elasticsearch.helpers import scan, bulk
from pydantic import BaseSettings
from rubrix.server.tasks.commons import TaskType


class Settings(BaseSettings):
    """
    Migration argument settings
    """

    elasticsearch: str = "http://localhost:9200"
    migration_datasets: List[str] = []
    chunk_size: int = 1000
    task: TaskType


settings = Settings()


source_datasets_index = ".rubric.datasets-v1"
target_datasets_index = ".rubrix.datasets-v0"
source_record_index_pattern = ".rubric.dataset.{}.records-v1"
target_record_index_pattern = ".rubrix.dataset.{}.records-v0"


def batcher(iterable, n, fillvalue=None):
    "batches an iterable"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def map_doc_2_action(
    index: str, doc: Dict[str, Any], task: TaskType
) -> Optional[Dict[str, Any]]:
    """Configures bulk action"""
    doc_data = doc["_source"]
    new_record = {
        "id": doc_data["id"],
        "metadata": doc_data.get("metadata"),
        "last_updated": doc_data.get("last_updated"),
        "words": doc_data.get("words"),
    }

    task_info = doc_data["tasks"].get(task)
    if task_info is None:
        return None

    new_record.update(
        {
            "status": task_info.get("status"),
            "prediction": task_info.get("prediction"),
            "annotation": task_info.get("annotation"),
            "event_timestamp": task_info.get("event_timestamp"),
            "predicted": task_info.get("predicted"),
            "annotated_as": task_info.get("annotated_as"),
            "predicted_as": task_info.get("predicted_as"),
            "annotated_by": task_info.get("annotated_by"),
            "predicted_by": task_info.get("predicted_by"),
            "score": task_info.get("confidences"),
            "owner": task_info.get("owner"),
        }
    )

    if task == TaskType.text_classification:
        new_record.update(
            {
                "inputs": doc_data.get("text"),
                "multi_label": task_info.get("multi_label"),
                "explanation": task_info.get("explanation"),
            }
        )
    elif task == TaskType.token_classification:
        new_record.update(
            {
                "tokens": doc_data.get("tokens"),
                "text": doc_data.get("raw_text"),
            }
        )
    return {
        "_op_type": "index",
        "_index": index,
        "_id": doc["_id"],
        **new_record,
    }


if __name__ == "__main__":

    client = Elasticsearch(hosts=settings.elasticsearch)

    for dataset in settings.migration_datasets:
        source_index = source_record_index_pattern.format(dataset)
        source_index_info = client.get(index=source_datasets_index, id=dataset)

        target_dataset_name = f"{dataset}-{settings.task}".lower()
        target_index = target_record_index_pattern.format(
            target_dataset_name
        )

        target_index_info = source_index_info["_source"]
        target_index_info["task"] = settings.task
        target_index_info["name"] = target_dataset_name

        client.index(
            index=target_datasets_index,
            id=target_index_info["name"],
            body=target_index_info,
        )

        index_docs = scan(client, index=source_index)
        for batch in batcher(index_docs, n=settings.chunk_size):
            bulk(
                client,
                actions=(
                    map_doc_2_action(index=target_index, doc=doc, task=settings.task)
                    for doc in batch
                    if doc is not None
                ),
            )



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import ClassVar, List, Optional

from fastapi import Depends

from rubrix.server.commons.errors import (
    EntityAlreadyExistsError,
    EntityNotFoundError,
    ForbiddenOperationError,
    WrongTaskError,
)
from .dao import DatasetsDAO, create_datasets_dao
from .model import (
    CopyDatasetRequest,
    CreationDatasetRequest,
    Dataset,
    DatasetDB,
    TaskType,
    UpdateDatasetRequest,
)
from ..security.model import User


class DatasetsService:
    """Datasets service"""

    _INSTANCE: ClassVar["DatasetsService"] = None

    @classmethod
    def get_instance(
        cls, dao: DatasetsDAO = Depends(create_datasets_dao)
    ) -> "DatasetsService":

        """
        Creates an instance of service

        Parameters
        ----------
        dao:
            The datasets dao

        Returns
        -------
            An instance of dataset service

        """

        if not cls._INSTANCE:
            cls._INSTANCE = cls(dao)
        return cls._INSTANCE

    def __init__(self, dao: DatasetsDAO):
        self.__dao__ = dao

    def list(
        self, user: User, workspaces: List[str], task_type: Optional[TaskType] = None
    ) -> List[Dataset]:
        """
        List datasets for a list of owners and task types

        Parameters
        ----------
        user:
            The request user
        workspaces:
            A list of selected user teams
        task_type:
            The task type: Optional

        Returns
        -------
            A list of datasets
        """
        owners = user.check_workspaces(workspaces)

        return [
            Dataset(**obs.dict())
            for obs in self.__dao__.list_datasets(owner_list=owners)
            if task_type is None or task_type == obs.task
        ]

    def create(
        self,
        dataset: CreationDatasetRequest,
        task: TaskType,
        user: User,
        workspace: Optional[str],
    ) -> Dataset:
        """
        Creates a datasets from given creation request

        Parameters
        ----------
        dataset:
            The dataset creation fields
        task:
            The dataset task
        user:
            The current user
        workspace:
            The user workspace selected for dataset creation

        Returns
        -------
            The created dataset

        """
        owner = user.check_workspace(workspace)
        date_now = datetime.utcnow()
        db_dataset = DatasetDB(
            **dataset.dict(by_alias=True),
            task=task,
            owner=owner,
            created_at=date_now,
            last_updated=date_now,
        )
        created_dataset = self.__dao__.create_dataset(db_dataset)
        return Dataset.parse_obj(created_dataset)

    def find_by_name(
        self, name: str, task: Optional[TaskType], user: User, workspace: Optional[str]
    ) -> DatasetDB:
        """
        Find a dataset by name

        Parameters
        ----------
        name:
            The dataset name
        task:
            Related dataset task
        user:
            The current user
        workspace:
            The user workspace where dataset belongs to

        Returns
        -------
            - The found dataset
            - EntityNotFoundError if not found
            - ForbiddenOperationError if user cannot access the dataset

        """
        owner = user.check_workspace(workspace)
        found = self.__dao__.find_by_name(name, owner=owner)
        if not found:
            raise EntityNotFoundError(name=name, type=Dataset)
        if found.owner and owner and found.owner != owner:
            raise ForbiddenOperationError()
        if task and found.task != task:
            raise WrongTaskError(f"Provided task {task} cannot be applied to dataset")
        return found

    def delete(self, name: str, user: User, workspace: Optional[str]):
        """
        Deletes a dataset.

        Parameters
        ----------
        name:
            The dataset name
        user:
            The current user
        workspace:
            The team where dataset belongs to

        """
        owner = user.check_workspace(workspace)
        found = self.__dao__.find_by_name(name, owner)
        if found:
            self.__dao__.delete_dataset(found)

    def update(
        self,
        name: str,
        data: UpdateDatasetRequest,
        user: User,
        workspace: Optional[str],
    ) -> Dataset:
        """
        Updates an existing dataset. Fields in update data are
        merged with store ones. Updates cannot remove data fields.

        Parameters
        ----------
        name:
            The dataset name
        data:
            The update fields
        user:
            The current user
        workspace:
            The workspace where dataset belongs to

        Returns
        -------
            The updated dataset
        """

        found = self.find_by_name(name, task=None, user=user, workspace=workspace)

        data.tags = {**found.tags, **data.tags}
        data.metadata = {**found.metadata, **data.metadata}
        updated = found.copy(
            update={**data.dict(by_alias=True), "last_updated": datetime.utcnow()}
        )

        self.__dao__.update_dataset(updated)
        return Dataset(**updated.dict(by_alias=True))

    def upsert(
        self,
        dataset: CreationDatasetRequest,
        task: TaskType,
        user: User,
        workspace: Optional[str],
    ) -> Dataset:
        """
        Inserts or updates the dataset. Updates only affects to updatable fields

        Parameters
        ----------
        dataset:
            The dataset data
        task:
            Selected task when dataset does not exists. Won't be used
            if dataset already exists
        user:
            The current user
        workspace:
            The dataset where dataset belongs to

        Returns
        -------

            The updated or created dataset

        """
        try:
            return self.update(
                name=dataset.name,
                data=UpdateDatasetRequest(tags=dataset.tags, metadata=dataset.metadata),
                user=user,
                workspace=workspace,
            )
        except EntityNotFoundError:
            return self.create(
                dataset=dataset, task=task, user=user, workspace=workspace
            )

    def copy_dataset(
        self,
        name: str,
        data: CopyDatasetRequest,
        user: User,
        workspace: Optional[str],
    ) -> Dataset:
        """
        Copies a dataset into another

        Parameters
        ----------
        name:
            The source dataset name
        data:
            A copy request configuration
        user:
            The current user
        workspace:
            The workspace where source dataset belongs to

        Returns
        -------

        """
        dataset_workspace = data.target_workspace or workspace
        dataset_workspace = user.check_workspace(dataset_workspace)

        try:
            self.find_by_name(data.name, task=None, user=user, workspace=dataset_workspace)
            raise EntityAlreadyExistsError(name=data.name, type=Dataset, workspace=dataset_workspace)
        except (EntityNotFoundError, ForbiddenOperationError):
            pass

        found = self.find_by_name(name, task=None, user=user, workspace=workspace)
        date_now = datetime.utcnow()

        created_dataset = DatasetDB(
            name=data.name,
            task=found.task,
            owner=dataset_workspace,
            created_at=date_now,
            last_updated=date_now,
            tags={**found.tags, **data.tags},
            metadata={
                **found.metadata,
                **data.metadata,
                "source_workspace": workspace,
                "copied_from": found.name,
            },
        )
        self.__dao__.copy(
            source=found,
            target=created_dataset,
        )

        return Dataset.parse_obj(created_dataset)

    def close_dataset(self, name: str, user: User, workspace: Optional[str]):
        """
        Closes a dataset. That means that all related dataset resources
        will be releases, but dataset cannot be explored

        Parameters
        ----------
        name:
            The dataset name
        user:
            The current user
        workspace:
            The workspace where dataset belongs to

        """
        found = self.find_by_name(name, task=None, user=user, workspace=workspace)
        self.__dao__.close(found)

    def open_dataset(self, name: str, user: User, workspace: Optional[str]):
        """
        Open a dataset. That means that all related dataset resources
        will be loaded and dataset will be ready for searches

        Parameters
        ----------
        name:
            The dataset name
        user:
            The current user
        workspace:
            The workspace where dataset belongs to

        """
        found = self.find_by_name(name, task=None, user=user, workspace=workspace)
        self.__dao__.open(found)



from rubrix.client.sdk.users.models import User
from rubrix.server.security.model import User as ServerUser


def test_users_schema(helpers):
    client_schema = User.schema()
    server_schema = ServerUser.schema()

    for clean_method in [helpers.remove_description, helpers.remove_pattern]:
        client_schema = clean_method(client_schema)
        server_schema = clean_method(server_schema)

    assert client_schema == server_schema



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict

from pydantic import BaseModel, Field

from rubrix._constants import API_KEY_HEADER_NAME


class Client(BaseModel):
    base_url: str
    cookies: Dict[str, str] = Field(default_factory=dict)
    headers: Dict[str, str] = Field(default_factory=dict)
    timeout: float = 5.0

    def get_headers(self) -> Dict[str, str]:
        """Get headers to be used in all endpoints"""
        return {**self.headers}

    def get_cookies(self) -> Dict[str, str]:
        return {**self.cookies}

    def get_timeout(self) -> float:
        return self.timeout


class AuthenticatedClient(Client):
    """A Client which has been authenticated for use on secured endpoints"""

    token: str

    def get_headers(self) -> Dict[str, str]:
        """Get headers to be used in authenticated endpoints"""
        return {
            "Authorization": f"Bearer {self.token}",  # Backward compatibility
            API_KEY_HEADER_NAME: self.token,
            **super().get_headers(),
        }

    def __hash__(self):
        return hash((self.base_url, self.token))



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import time
from typing import Any, Dict

import rubrix
from fastapi import FastAPI
from rubrix import TextClassificationRecord, TokenClassificationRecord
from rubrix.monitoring.asgi import RubrixLogHTTPMiddleware, token_classification_mapper
from starlette.applications import Starlette
from starlette.responses import JSONResponse, PlainTextResponse
from starlette.testclient import TestClient


def test_rubrix_middleware_for_text_classification(monkeypatch):

    expected_endpoint = "/predict"
    expected_dataset_name = "mlmodel_v3_monitor_ds"

    app = FastAPI()
    app.add_middleware(
        RubrixLogHTTPMiddleware,
        api_endpoint=expected_endpoint,
        dataset=expected_dataset_name,
    )

    @app.route(expected_endpoint, methods=["POST"])
    def mock_predict(data: Dict[str, Any]):
        return JSONResponse(
            content=[
                {"labels": ["A", "B"], "scores": [0.9, 0.1]},
                {"labels": ["A", "B"], "scores": [0.9, 0.1]},
            ]
        )

    @app.route("/another/predict/route")
    def another_mock(request):
        return PlainTextResponse("Hello")

    class MockLog:
        def __init__(self):
            self.was_called = False

        def __call__(self, records, name: str, **kwargs):
            self.was_called = True
            assert name == expected_dataset_name
            assert len(records) == 2
            assert isinstance(records[0], TextClassificationRecord)

    mock_log = MockLog()
    monkeypatch.setattr(rubrix, "log", mock_log)
    mock = TestClient(app)

    mock.post(
        expected_endpoint,
        json=[
            {"a": "The data input for A", "b": "The data input for B"},
            {"a": "The data input for A", "b": "The data input for B"},
        ],
    )

    assert mock_log.was_called
    time.sleep(0.200)
    mock_log.was_called = False
    mock.get("/another/predict/route")
    assert not mock_log.was_called


def test_rubrix_middleware_for_token_classification(monkeypatch):

    expected_endpoint = "/predict"
    expected_dataset_name = "mlmodel_v3_monitor_ds"

    app = Starlette()
    app.add_middleware(
        RubrixLogHTTPMiddleware,
        api_endpoint=expected_endpoint,
        dataset=expected_dataset_name,
        records_mapper=token_classification_mapper,
    )

    @app.route(expected_endpoint, methods=["POST"])
    def mock_predict(request):
        return JSONResponse(
            content=[
                [
                    {"label": "fawn", "start": 1, "end": 10},
                    {"label": "fobis", "start": 12, "end": 14},
                ],
                [],
            ]
        )

    class MockLog:
        def __init__(self):
            self.was_called = False

        def __call__(self, records, name: str, **kwargs):
            self.was_called = True
            assert name == expected_dataset_name
            assert len(records) == 2
            assert isinstance(records[0], TokenClassificationRecord)

    mock_log = MockLog()
    monkeypatch.setattr(rubrix, "log", mock_log)
    mock = TestClient(app)

    mock.post(
        expected_endpoint,
        json=[{"text": "The main text data"}, {"text": "The main text data"}],
    )
    time.sleep(0.2)
    assert mock_log.was_called

    mock_log.was_called = False
    mock.get("/another/predict/route")

    time.sleep(0.2)
    assert not mock_log.was_called



def start_loop_in_thread():
    """Launches a asyncio loop in a different thread and start it"""
    from threading import Thread
    import asyncio

    def start_loop(loop):
        asyncio.set_event_loop(loop)
        loop.run_forever()

    new_loop = asyncio.new_event_loop()
    t = Thread(target=start_loop, args=(new_loop,), daemon=True)
    t.start()

    return new_loop



import pytest
from pydantic import ValidationError

from rubrix.server.commons.errors import ForbiddenOperationError
from rubrix.server.security.model import User


@pytest.mark.parametrize("email", ["my@email.com", "infra@recogn.ai"])
def test_valid_mail(email):
    user = User(username="user", email=email)
    assert user.email == email


@pytest.mark.parametrize(
    "wrong_email", ["non-valid-email", "wrong@mail", "@wrong" "wrong.mail"]
)
def test_email_validator(wrong_email):
    with pytest.raises(ValidationError):
        User(username="user", email=wrong_email)


@pytest.mark.parametrize("wrong_name", ["user name", "user/name", "user.name"])
def test_username_validator(wrong_name):
    with pytest.raises(ValidationError):
        User(username=wrong_name)


@pytest.mark.parametrize("wrong_workspace", ["work space", "work/space", "work.space"])
def test_workspace_validator(wrong_workspace):
    with pytest.raises(ValidationError):
        User(username="username", workspaces=[wrong_workspace])


def test_check_non_provided_workspaces():
    user = User(username="test")
    assert not user.check_workspaces([])  # super-user

    user.workspaces = ["ws"]
    assert user.check_workspaces([]) == [user.default_workspace]

    with pytest.raises(ForbiddenOperationError, match="not-found"):
        assert user.check_workspaces(["ws", "not-found"])


def test_check_user_workspaces():

    a_ws = "A-workspace"
    expected_workspaces = [a_ws, "B-ws"]
    user = User(username="test-user", workspaces=[a_ws, "B-ws", "C-ws"])

    assert user.check_workspace(a_ws) == a_ws
    assert user.check_workspaces(expected_workspaces) == expected_workspaces
    with pytest.raises(ForbiddenOperationError):
        assert user.check_workspaces(["not-found-ws"])


def test_default_workspace():

    user = User(username="admin")
    assert user.default_workspace is None

    test_user = User(username="test", workspaces=["ws"])
    assert test_user.default_workspace == test_user.username


def test_workspace_for_superuser():
    user = User(username="admin")
    assert user.default_workspace is None

    assert user.check_workspace("some") == "some"
    assert user.check_workspaces(["some"]) == ["some"]

@pytest.mark.parametrize(
    "workspaces, expected",
    [
        (None, []),
        ([], ["user"]),
        (["a"], ["user"]),
    ],
)
def test_check_team_with_default(workspaces, expected):
    user = User(username="user", workspaces=workspaces)
    assert user.check_workspaces([]) == expected
    assert user.check_workspaces(None) == expected
    assert user.check_workspaces([None]) == expected



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
This module configures the api routes under /api prefix, and
set the required security dependencies if api security is enabled
"""

from fastapi import APIRouter
from rubrix.server.commons.errors import ErrorMessage

from .datasets import api as datasets
from .info import api as info
from .tasks import api as tasks
from .users import api as users

api_router = APIRouter(
    responses={
        "404": {"model": ErrorMessage, "description": "Item not found"},
        "500": {"model": ErrorMessage, "description": "Generic error"},
    }
)


dependencies = []

for router in [users.router, datasets.router, info.router, tasks.router]:
    api_router.include_router(router, dependencies=dependencies)



from typing import Optional

from pydantic import BaseModel, Field


class MetricInfo(BaseModel):
    """Metric info data model for retrieve dataset metrics information"""

    id: str = Field(description="The metric id")
    name: str = Field(description="The metric name")
    description: Optional[str] = Field(
        default=None, description="The metric description"
    )



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import itertools
from typing import Iterable, Optional

from fastapi import APIRouter, Depends, Query, Security
from fastapi.responses import StreamingResponse

from rubrix.server.commons.api import CommonTaskQueryParams
from rubrix.server.datasets.model import CreationDatasetRequest, Dataset
from rubrix.server.datasets.service import DatasetsService
from rubrix.server.security import auth
from rubrix.server.security.model import User
from rubrix.server.tasks.commons import (
    BulkResponse,
    PaginationParams,
    TaskType,
)
from rubrix.server.tasks.commons.helpers import takeuntil
from rubrix.server.tasks.token_classification.api.model import (
    TokenClassificationBulkData,
    TokenClassificationQuery,
    TokenClassificationRecord,
    TokenClassificationSearchRequest,
    TokenClassificationSearchResults,
)
from rubrix.server.tasks.token_classification.service.service import (
    TokenClassificationService,
    token_classification_service,
)

TASK_TYPE = TaskType.token_classification
BASE_ENDPOINT = "/{name}/" + TASK_TYPE

router = APIRouter(tags=[TASK_TYPE], prefix="/datasets")


@router.post(
    BASE_ENDPOINT + ":bulk",
    operation_id="bulk_records",
    response_model=BulkResponse,
    response_model_exclude_none=True,
)
def bulk_records(
    name: str,
    bulk: TokenClassificationBulkData,
    common_params: CommonTaskQueryParams = Depends(),
    service: TokenClassificationService = Depends(token_classification_service),
    datasets: DatasetsService = Depends(DatasetsService.get_instance),
    current_user: User = Security(auth.get_user, scopes=[]),
) -> BulkResponse:
    """
    Includes a chunk of record data with provided dataset bulk information

    Parameters
    ----------
    name:
        The dataset name
    bulk:
        The bulk data
    common_params:
        Common query params
    service:
        the Service
    datasets:
        The dataset service
    current_user:
        Current request user

    Returns
    -------
        Bulk response data
    """

    dataset = datasets.upsert(
        CreationDatasetRequest(**{**bulk.dict(), "name": name}),
        user=current_user,
        workspace=common_params.workspace,
        task=TASK_TYPE,
    )
    result = service.add_records(
        dataset=dataset,
        records=bulk.records,
    )
    return BulkResponse(
        dataset=name,
        processed=result.processed,
        failed=result.failed,
    )


@router.post(
    BASE_ENDPOINT + ":search",
    response_model=TokenClassificationSearchResults,
    response_model_exclude_none=True,
    operation_id="search_records",
)
def search_records(
    name: str,
    search: TokenClassificationSearchRequest = None,
    common_params: CommonTaskQueryParams = Depends(),
    pagination: PaginationParams = Depends(),
    service: TokenClassificationService = Depends(token_classification_service),
    datasets: DatasetsService = Depends(DatasetsService.get_instance),
    current_user: User = Security(auth.get_user, scopes=[]),
) -> TokenClassificationSearchResults:
    """
    Searches data from dataset

    Parameters
    ----------
    name:
        The dataset name
    search:
        THe search query request
    common_params:
        Common query params
    pagination:
        The pagination params
    service:
        The dataset records service
    datasets:
        The dataset service
    current_user:
        The current request user

    Returns
    -------
        The search results data

    """

    search = search or TokenClassificationSearchRequest()
    query = search.query or TokenClassificationQuery()

    dataset = datasets.find_by_name(
        name, task=TASK_TYPE, user=current_user, workspace=common_params.workspace
    )
    result = service.search(
        dataset=Dataset.parse_obj(dataset),
        query=query,
        sort_by=search.sort,
        record_from=pagination.from_,
        size=pagination.limit,
    )

    return result


def scan_data_response(
    data_stream: Iterable[TokenClassificationRecord],
    chunk_size: int = 1000,
    limit: Optional[int] = None,
) -> StreamingResponse:
    """Generate an textual stream data response for a dataset scan"""

    async def stream_generator(stream):
        """Converts dataset scan into a text stream"""

        def grouper(n, iterable, fillvalue=None):
            args = [iter(iterable)] * n
            return itertools.zip_longest(fillvalue=fillvalue, *args)

        if limit:
            stream = takeuntil(stream, limit=limit)

        for batch in grouper(
            n=chunk_size,
            iterable=stream,
        ):
            filtered_records = filter(lambda r: r is not None, batch)
            yield "\n".join(
                map(
                    lambda r: r.json(by_alias=True, exclude_none=True), filtered_records
                )
            ) + "\n"

    return StreamingResponse(
        stream_generator(data_stream), media_type="application/json"
    )


@router.post(
    BASE_ENDPOINT + "/data",
    operation_id="stream_data",
)
async def stream_data(
    name: str,
    query: Optional[TokenClassificationQuery] = None,
    common_params: CommonTaskQueryParams = Depends(),
    limit: Optional[int] = Query(None, description="Limit loaded records", gt=0),
    service: TokenClassificationService = Depends(token_classification_service),
    datasets: DatasetsService = Depends(DatasetsService.get_instance),
    current_user: User = Security(auth.get_user, scopes=[]),
) -> StreamingResponse:
    """
    Creates a data stream over dataset records

    Parameters
    ----------
    name
        The dataset name
    query:
        The stream data query
    common_params:
        Common query params
    limit:
        The load number of records limit. Optional
    service:
        The dataset records service
    datasets:
        The datasets service
    current_user:
        Request user

    """
    query = query or TokenClassificationQuery()
    dataset = datasets.find_by_name(
        name, task=TASK_TYPE, user=current_user, workspace=common_params.workspace
    )
    data_stream = service.read_dataset(dataset=Dataset.parse_obj(dataset), query=query)

    return scan_data_response(
        data_stream=data_stream,
        limit=limit,
    )



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import socket
from datetime import datetime

import pytest

from rubrix.client.models import TextClassificationRecord, TokenAttributions
from rubrix.client.sdk.text_classification.models import (
    ClassPrediction,
    CreationTextClassificationRecord,
    TextClassificationAnnotation,
    TextClassificationBulkData,
    TextClassificationQuery,
)
from rubrix.client.sdk.text_classification.models import (
    TextClassificationRecord as SdkTextClassificationRecord,
)
from rubrix.server.tasks.text_classification.api.model import (
    TextClassificationBulkData as ServerTextClassificationBulkData,
)
from rubrix.server.tasks.text_classification.api.model import (
    TextClassificationQuery as ServerTextClassificationQuery,
)


def test_bulk_data_schema(helpers):
    client_schema = TextClassificationBulkData.schema()
    server_schema = ServerTextClassificationBulkData.schema()

    assert helpers.remove_description(client_schema) == helpers.remove_description(
        server_schema
    )


def test_query_schema(helpers):
    client_schema = TextClassificationQuery.schema()
    server_schema = ServerTextClassificationQuery.schema()

    assert helpers.remove_description(client_schema) == helpers.remove_description(
        server_schema
    )


def test_from_client_explanation():
    token_attributions = [
        TokenAttributions(token="test", attributions={"label1": 1.0, "label2": 2.0})
    ]
    record = TextClassificationRecord(
        inputs={"text": "test"},
        prediction=[("label1", 0.5), ("label2", 0.5)],
        annotation="label1",
        explanation={"text": token_attributions},
        event_timestamp=datetime(2000, 1, 1),
        id=1,
    )
    sdk_record = CreationTextClassificationRecord.from_client(record)

    assert sdk_record.explanation["text"] == token_attributions


@pytest.mark.parametrize(
    "annotation,expected", [("label1", 1), (["label1", "label2"], 2)]
)
def test_from_client_annotation(annotation, expected):
    record = TextClassificationRecord(
        inputs={"text": "test"},
        annotation=annotation,
        multi_label=True,
    )
    sdk_record = CreationTextClassificationRecord.from_client(record)

    assert len(sdk_record.annotation.labels) == expected


@pytest.mark.parametrize(
    "pred_agent,annot_agent,pred_expected,annot_expected",
    [
        (None, None, socket.gethostname(), socket.gethostname()),
        ("pred_agent", "annot_agent", "pred_agent", "annot_agent"),
    ],
)
def test_from_client_agent(pred_agent, annot_agent, pred_expected, annot_expected):
    record = TextClassificationRecord(
        inputs="test",
        prediction=[("label1", 0.5), ("label2", 0.5)],
        annotation="label1",
        prediction_agent=pred_agent,
        annotation_agent=annot_agent,
        metrics={"not_passed_on": 0},
    )
    sdk_record = CreationTextClassificationRecord.from_client(record)

    assert sdk_record.annotation.agent == annot_expected
    assert sdk_record.prediction.agent == pred_expected
    assert sdk_record.metrics == {}


@pytest.mark.parametrize(
    "multi_label,expected", [(False, "annot_label"), (True, ["annot_label"])]
)
def test_to_client(multi_label, expected):
    annotation = TextClassificationAnnotation(
        labels=[ClassPrediction(**{"class": "annot_label"})], agent="annot_agent"
    )
    prediction = TextClassificationAnnotation(
        labels=[
            ClassPrediction(**{"class": "label1", "score": 0.5}),
            ClassPrediction(**{"class": "label2", "score": 0.5}),
        ],
        agent="pred_agent",
    )

    sdk_record = SdkTextClassificationRecord(
        inputs={"text": "test"},
        annotation=annotation,
        prediction=prediction,
        multi_label=multi_label,
        event_timestamp=datetime(2000, 1, 1),
        metrics={"tokens_length": 42},
    )

    record = sdk_record.to_client()

    assert record.prediction == [("label1", 0.5), ("label2", 0.5)]
    assert record.prediction_agent == "pred_agent"
    assert record.annotation_agent == "annot_agent"
    assert record.annotation == expected
    assert record.metrics == {"tokens_length": 42}



import pytest

from rubrix.server.security.auth_provider.local.users.service import create_users_dao

usersDAO = create_users_dao()


# Test for function get_user class UserInDB
def test_get_user():
    user = usersDAO.get_user(user_name="rubrix")
    assert user.username == "rubrix"


# Test for function get_user_by_api_key class UserInDB
@pytest.mark.asyncio
async def test_get_user_by_api_key():
    user = await usersDAO.get_user_by_api_key(api_key="rubrix.apikey")
    assert user.username == "rubrix"



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from fastapi import APIRouter

from rubrix.server.tasks.token_classification.metrics import TokenClassificationMetrics
from .commons import TaskType
from .commons.metrics.api import configure_metrics_endpoints
from .commons.task_factory import TaskFactory
from .text2text import Text2TextQuery, Text2TextRecord, api as text2text
from .text2text.metrics import Text2TextMetrics
from .text_classification import (
    TextClassificationQuery,
    TextClassificationRecord,
    api as text_classification,
)
from .text_classification.metrics import TextClassificationMetrics
from .token_classification import (
    TokenClassificationQuery,
    TokenClassificationRecord,
    api as token_classification,
)

router = APIRouter()

TaskFactory.register_task(
    task_type=TaskType.token_classification,
    query_request=TokenClassificationQuery,
    record_class=TokenClassificationRecord,
    metrics=TokenClassificationMetrics,
)

TaskFactory.register_task(
    task_type=TaskType.text_classification,
    query_request=TextClassificationQuery,
    record_class=TextClassificationRecord,
    metrics=TextClassificationMetrics,
)

TaskFactory.register_task(
    task_type=TaskType.text2text,
    query_request=Text2TextQuery,
    record_class=Text2TextRecord,
    metrics=Text2TextMetrics,
)


for task_api in [text_classification, token_classification, text2text]:
    cfg = TaskFactory.get_task_by_task_type(task_api.TASK_TYPE)
    if cfg:
        configure_metrics_endpoints(task_api.router, cfg)

    router.include_router(task_api.router)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Common helper functions
"""

from typing import Any, Dict, List, Optional


def flatten_dict(
    data: Dict[str, Any], sep: str = ".", drop_empty: bool = False
) -> Dict[str, Any]:
    """
    Flatten a data dictionary

    Parameters
    ----------
    data:
        The data dictionary
    sep:
        The generated key separator. Default="."
    drop_empty:
        If true, keys with empty lists or None values will be omitted

    Returns
    -------

        A flattened key dictionary
    """

    def _is_empty_value(value: Any) -> bool:
        if value is None:
            return True
        if isinstance(value, list) and len(value) == 0:
            return True
        return False

    def _flatten_internal_(_data: Dict[str, Any], _parent_key="", _sep="."):
        items = []
        for key, value in _data.items():
            if drop_empty and _is_empty_value(value):
                continue

            new_key = _parent_key + _sep + key if _parent_key else key
            try:
                items.extend(_flatten_internal_(value, new_key, _sep=_sep).items())
            except Exception:
                items.append((new_key, value))
        return dict(items)

    return _flatten_internal_(data, _sep=sep)


def unflatten_dict(
    data: Dict[str, Any], sep: str = ".", stop_keys: Optional[List[str]] = None
) -> Dict[str, Any]:
    """
    Given a flat dictionary keys, build a hierarchical version by grouping keys

    Parameters
    ----------
    data:
        The data dictionary
    sep:
        The key separator. Default "."
    stop_keys
        List of dictionary first level keys where hierarchy will stop

    Returns
    -------

    """
    resultDict = {}

    for key, value in data.items():
        if key is not None:
            parts = key.split(sep)
            if parts[0] in stop_keys:
                parts = [parts[0], sep.join(parts[1:])]
            d = resultDict
            for part in parts[:-1]:
                if part not in d:
                    d[part] = {}
                d = d[part]
            d[parts[-1]] = value
    return resultDict


def limit_value_length(data: Any, max_length: int) -> Any:
    """
    Given an input data, limits string values to a max_length by fetching
    last max_length characters

    Parameters
    ----------
    data:
        Input data
    max_length:
        Max length for string values

    Returns
    -------
        Limited version of data, if any
    """

    if isinstance(data, str):
        return data[-max_length:]
    if isinstance(data, dict):
        return {
            k: limit_value_length(v, max_length=max_length) for k, v in data.items()
        }
    if isinstance(data, (list, tuple, set)):
        new_values = map(lambda x: limit_value_length(x, max_length=max_length), data)
        return type(data)(new_values)
    return data



import httpx

from rubrix.client.sdk.client import AuthenticatedClient
from rubrix.client.sdk.commons.models import ErrorMessage, HTTPValidationError, Response
from rubrix.client.sdk.users.models import User


def _parse_response(response: httpx.Response):
    parsed_response = None
    if response.status_code == 200:
        parsed_response = User(**response.json())
    elif response.status_code == 404:
        parsed_response = ErrorMessage(**response.json())
    elif response.status_code == 500:
        parsed_response = ErrorMessage(**response.json())
    elif response.status_code == 422:
        parsed_response = HTTPValidationError(**response.json())

    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=parsed_response,
    )


def whoami(client: AuthenticatedClient):
    url = "{}/api/me".format(client.base_url)

    response = httpx.get(
        url=url,
        headers=client.get_headers(),
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
    )

    return _parse_response(response)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Iterable, List, Optional

from fastapi import Depends

from rubrix.server.commons.es_helpers import (
    aggregations,
    sort_by2elasticsearch,
)
from rubrix.server.datasets.model import Dataset
from rubrix.server.tasks.commons import (
    BulkResponse,
    EsRecordDataFieldNames,
    SortableField,
)
from rubrix.server.tasks.commons.dao import extends_index_properties
from rubrix.server.tasks.commons.dao.dao import DatasetRecordsDAO, dataset_records_dao
from rubrix.server.tasks.commons.dao.model import RecordSearch
from rubrix.server.tasks.commons.metrics.service import MetricsService
from rubrix.server.tasks.text2text.api.model import (
    CreationText2TextRecord,
    ExtendedEsRecordDataFieldNames,
    Text2TextQuery,
    Text2TextRecord,
    Text2TextRecordDB,
    Text2TextSearchAggregations,
    Text2TextSearchResults,
)

extends_index_properties(
    {
        ExtendedEsRecordDataFieldNames.text_predicted: {
            "type": "text",
            "fielddata": True,
            "analyzer": "multilingual_stop_analyzer",
            "fields": {"extended": {"type": "text", "analyzer": "extended_analyzer"}},
        },
        ExtendedEsRecordDataFieldNames.text_annotated: {
            "type": "text",
            "fielddata": True,
            "analyzer": "multilingual_stop_analyzer",
            "fields": {"extended": {"type": "text", "analyzer": "extended_analyzer"}},
        },
    }
)


class Text2TextService:
    """
    Text2text service

    """

    def __init__(
        self,
        dao: DatasetRecordsDAO,
        metrics: MetricsService,
    ):
        self.__dao__ = dao
        self.__metrics__ = metrics

    def add_records(
        self,
        dataset: Dataset,
        records: List[CreationText2TextRecord],
    ):
        self.__metrics__.build_records_metrics(dataset, records)
        failed = self.__dao__.add_records(
            dataset=dataset,
            records=records,
            record_class=Text2TextRecordDB,
        )
        return BulkResponse(dataset=dataset.name, processed=len(records), failed=failed)

    def search(
        self,
        dataset: Dataset,
        query: Text2TextQuery,
        sort_by: List[SortableField],
        record_from: int = 0,
        size: int = 100,
    ) -> Text2TextSearchResults:
        """
        Run a search in a dataset

        Parameters
        ----------
        dataset:
            The records dataset
        query:
            The search parameters
        sort_by:
            The sort by list
        record_from:
            The record from return results
        size:
            The max number of records to return

        Returns
        -------
            The matched records with aggregation info for specified task_meta.py

        """
        results = self.__dao__.search_records(
            dataset,
            search=RecordSearch(
                query=query.as_elasticsearch(),
                sort=sort_by2elasticsearch(
                    sort_by,
                    valid_fields=[
                        "metadata",
                        EsRecordDataFieldNames.score,
                        EsRecordDataFieldNames.predicted,
                        EsRecordDataFieldNames.predicted_as,
                        EsRecordDataFieldNames.predicted_by,
                        EsRecordDataFieldNames.annotated_as,
                        EsRecordDataFieldNames.annotated_by,
                        EsRecordDataFieldNames.status,
                        EsRecordDataFieldNames.event_timestamp,
                    ],
                ),
                aggregations={
                    ExtendedEsRecordDataFieldNames.text_predicted: aggregations.terms_aggregation(
                        ExtendedEsRecordDataFieldNames.text_predicted
                    )
                },
            ),
            size=size,
            record_from=record_from,
        )
        return Text2TextSearchResults(
            total=results.total,
            records=[Text2TextRecord.parse_obj(r) for r in results.records],
            aggregations=Text2TextSearchAggregations(
                **results.aggregations,
                words=results.words,
                metadata=results.metadata or {},
            )
            if results.aggregations
            else None,
        )

    def read_dataset(
        self,
        dataset: Dataset,
        query: Optional[Text2TextQuery] = None,
    ) -> Iterable[Text2TextRecord]:
        """
        Scan a dataset records

        Parameters
        ----------
        dataset:
            The records dataset
        query:
            If provided, scan will retrieve only records matching
            the provided query filters. Optional

        """
        for db_record in self.__dao__.scan_dataset(
            dataset, search=RecordSearch(query=query.as_elasticsearch())
        ):
            yield Text2TextRecord.parse_obj(db_record)


_instance = None


def text2text_service(
    dao: DatasetRecordsDAO = Depends(dataset_records_dao),
    metrics: MetricsService = Depends(MetricsService.get_instance),
) -> Text2TextService:
    """
    Creates a dataset record service instance

    Parameters
    ----------
    dao:
        The dataset records dao dependency
    metrics:
        The metrics service

    Returns
    -------
        A dataset records service instance
    """
    global _instance
    if not _instance:
        _instance = Text2TextService(dao=dao, metrics=metrics)
    return _instance



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import datetime
import json
import logging
import re
import threading
from queue import Queue
from typing import Any, Callable, Dict, List, Optional, Tuple

import rubrix
from rubrix import Record, TextClassificationRecord, TokenClassificationRecord
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import JSONResponse, Response, StreamingResponse
from starlette.types import Message, Receive

_logger = logging.getLogger(__name__)
_spaces_regex = re.compile(r"\s+")


def token_classification_mapper(inputs, outputs):
    text = inputs.get("text", "")
    tokens = outputs.get("tokens") if isinstance(outputs, dict) else None
    return TokenClassificationRecord(
        text=text,
        tokens=tokens or _spaces_regex.split(text),
        prediction=[
            (entity["label"], entity["start"], entity["end"])
            for entity in (
                outputs.get("entities") if isinstance(outputs, dict) else outputs
            )
        ],
        event_timestamp=datetime.datetime.now(),
    )


def text_classification_mapper(inputs, outputs):
    return TextClassificationRecord(
        inputs=inputs,
        prediction=[
            (label, score)
            for label, score in zip(
                outputs.get("labels", []), outputs.get("scores", [])
            )
        ],
        event_timestamp=datetime.datetime.now(),
    )


class CachedJsonRequest(Request):
    """
    We must a cached version of incoming requests since request body cannot be read from middleware directly.
    See <https://github.com/encode/starlette/issues/847> for more information

    TODO Remove usage of CachedRequest when https://github.com/encode/starlette/pull/848 is released
    """

    @property
    def receive(self) -> Receive:
        body = None
        if hasattr(self, "_body"):
            body = self._body
        if body is not None:

            async def cached_receive() -> Message:
                return dict(type="http.request", body=body)

            return cached_receive
        return self._receive


class RubrixLogHTTPMiddleware(BaseHTTPMiddleware):
    """An standard starlette middleware that enables rubrix logs for http prediction requests"""

    def __init__(
        self,
        api_endpoint: str,
        dataset: str,
        records_mapper: Optional[Callable[[dict, dict], Record]] = None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._endpoint = api_endpoint
        self._dataset = dataset
        self._records_mapper = records_mapper or text_classification_mapper
        self._queue = Queue()
        self._worker_task = threading.Thread(
            target=self.__worker__, name=RubrixLogHTTPMiddleware.__name__, daemon=True
        )
        self._worker_task.start()

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        if self._endpoint != request.url.path:  # Filtering endpoint path
            return await call_next(request)

        content_type = request.headers.get("Content-type", None)
        if "application/json" not in content_type:
            return await call_next(request)

        cached_request = CachedJsonRequest(
            scope=request.scope, receive=request.receive, send=request._send
        )
        # Must read body before call_next
        inputs = await cached_request.json()
        response: Response = await call_next(cached_request)
        try:
            if (
                not isinstance(response, (JSONResponse, StreamingResponse))
                or response.status_code >= 400
            ):
                return response

            new_response, outputs = await self._extract_response_content(response)
            self._queue.put_nowait((inputs, outputs, str(request.url)))
            return new_response
        except Exception as ex:
            _logger.error("Cannot log to rubrix", exc_info=ex)
            return response

    def __worker__(self):
        while True:
            try:
                inputs, outputs, url = self._queue.get()
                self._log_to_rubrix(inputs, outputs, url)
            except Exception as ex:
                # Run thread FOREVER!!!
                _logger.error("Error sending records to rubrix", exc_info=ex)
            finally:
                self._queue.task_done()

    async def _extract_response_content(
        self, response: Response
    ) -> Tuple[Response, List[Dict[str, Any]]]:
        """Extracts response body content from response and returns a new processable response"""
        body = b""
        new_response = response
        if isinstance(response, StreamingResponse):
            async for chunk in response.body_iterator:
                body += chunk
            new_response = StreamingResponse(
                content=(chunk for chunk in [body]),
                status_code=response.status_code,
                headers={k: v for k, v in response.headers.items()},
                media_type=response.media_type,
                background=response.background,
            )
        else:
            body = response.body
        return new_response, json.loads(body)

    def _log_to_rubrix(
        self,
        inputs: List[Dict[str, Any]],
        outputs: List[Dict[str, Any]],
        url: str,
        **tags
    ):
        records = [
            record
            for _inputs, _outputs in zip(inputs, outputs)
            for record in [self._records_mapper(_inputs, _outputs)]
            if record
        ]

        if records:
            for r in records:
                r.prediction_agent = url
            rubrix.log(records=records, name=self._dataset, tags=tags)



#  coding=utf-8
#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json
from typing import Any, Dict, List, Type, TypeVar, Union

import httpx

from rubrix.client.sdk.commons.models import (
    BulkResponse,
    ErrorMessage,
    HTTPValidationError,
    Response,
)


def build_bulk_response(
    response: httpx.Response,
) -> Response[Union[BulkResponse, ErrorMessage, HTTPValidationError]]:

    parsed_response = None
    if response.status_code == 200:
        parsed_response = BulkResponse(**response.json())
    elif response.status_code == 404:
        parsed_response = ErrorMessage(**response.json())
    elif response.status_code == 500:
        parsed_response = ErrorMessage(**response.json())
    elif response.status_code == 422:
        parsed_response = HTTPValidationError(**response.json())

    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=parsed_response,
    )


T = TypeVar("T")


def build_data_response(
    response: httpx.Response, data_type: Type[T]
) -> Response[Union[List[T], HTTPValidationError, ErrorMessage]]:
    if 200 <= response.status_code < 400:
        parsed_response = [data_type(**json.loads(r)) for r in response.iter_lines()]
    else:
        content = next(response.iter_lines())
        data = json.loads(content)
        parsed_response = (
            HTTPValidationError(**data)
            if response.status_code == 422
            else ErrorMessage(**data)
        )

    return Response(
        status_code=response.status_code,
        content=b"",
        headers=response.headers,
        parsed=parsed_response,
    )


def build_list_response(
    response: httpx.Response,
    item_class: Type[T],
) -> Response[Union[List[T], HTTPValidationError, ErrorMessage]]:
    parsed_response = response.json()

    if 200 <= response.status_code < 400:
        parsed_response = [item_class(**r) for r in parsed_response]
    elif response.status_code == 422:
        parsed_response = HTTPValidationError.parse_obj(parsed_response)
    else:
        parsed_response = ErrorMessage.parse_obj(parsed_response)

    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=parsed_response,
    )


def build_raw_response(
    response: httpx.Response,
) -> Response[Union[Dict[str, Any], ErrorMessage, HTTPValidationError]]:

    parsed_response = response.json()
    if response.status_code == 404:
        parsed_response = ErrorMessage(**parsed_response)
    elif response.status_code == 500:
        parsed_response = ErrorMessage(**parsed_response)
    elif response.status_code == 422:
        parsed_response = HTTPValidationError(**parsed_response)

    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=parsed_response,
    )



from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    TypeVar,
    Union,
)

from pydantic import BaseModel
from pydantic.generics import GenericModel

from rubrix.server.commons.es_helpers import aggregations
from rubrix.server.tasks.commons import BaseRecord


GenericRecord = TypeVar("GenericRecord", bound=BaseRecord)


class BaseMetric(BaseModel):
    """
    Base model for rubrix dataset metrics summaries
    """

    id: str
    name: str
    description: str = None


class PythonMetric(BaseMetric, Generic[GenericRecord]):
    """
    A metric definition which will be calculated using raw queried data
    """

    def apply(self, records: Iterable[GenericRecord]) -> Dict[str, Any]:
        """
        Metric calculation method.

        Parameters
        ----------
        records:
            The matched records

        Returns
        -------
            The metric result
        """
        raise NotImplementedError()


class ElasticsearchMetric(BaseMetric):
    """
    A metric summarized by using one or several elasticsearch aggregations
    """

    def aggregation_request(self, *args, **kwargs) -> Dict[str, Any]:
        """
        Configures the summary es aggregation definition
        """
        raise NotImplementedError()

    def aggregation_result(self, aggregation_result: Dict[str, Any]) -> Dict[str, Any]:
        """
        Parse the es aggregation result. Override this method
        for result customization

        Parameters
        ----------
        aggregation_result:
            Retrieved es aggregation result

        """
        return aggregation_result


class NestedPathElasticsearchMetric(ElasticsearchMetric):
    """
    A ``ElasticsearchMetric`` which need nested fields for summary calculation.

    Aggregations for nested fields need some extra configuration and this class
    encapsulate these common logic.

    Attributes:
    -----------
    nested_path:
        The nested
    """

    nested_path: str

    def inner_aggregation(self, *args, **kwargs) -> Dict[str, Any]:
        """The specific aggregation definition"""
        raise NotImplementedError()

    def aggregation_request(self, *args, **kwargs) -> Dict[str, Any]:
        """Implements the common mechanism to define aggregations with nested fields"""
        return {
            self.id: aggregations.nested_aggregation(
                nested_path=self.nested_path,
                inner_aggregation=self.inner_aggregation(*args, **kwargs),
            )
        }

    def compound_nested_field(self, inner_field: str) -> str:
        return f"{self.nested_path}.{inner_field}"


class BaseTaskMetrics(GenericModel, Generic[GenericRecord]):
    """
    Base class encapsulating related task metrics

    Attributes:
    -----------

    metrics:
        A list of configured metrics for task
    """

    metrics: ClassVar[List[BaseMetric]]

    @classmethod
    def configure_es_index(cls):
        """
        If some metrics require specific es field mapping definitions,
        include them here.

        """
        pass

    @classmethod
    def find_metric(cls, id: str) -> Optional[BaseMetric]:
        """
        Finds a metric by id

        Parameters
        ----------
        id:
            The metric id

        Returns
        -------
            Found metric if any, ``None`` otherwise

        """
        for metric in cls.metrics:
            if metric.id == id:
                return metric

    @classmethod
    def record_metrics(cls, record: GenericRecord) -> Dict[str, Any]:
        """
        Use this method is some configured metric requires additional
        records fields.

        Generated records will be persisted under ``metrics`` record path.
        For example, if you define a field called ``sentence_length`` like

        >>> def record_metrics(cls, record)-> Dict[str, Any]:
        ...     return { "sentence_length" : len(record.text) }

        The new field will be stored in elasticsearch in ``metrics.sentence_length``

        Parameters
        ----------
        record:
            The record used for calculate metrics fields

        Returns
        -------
            A dict with calculated metrics fields
        """
        return {}


class HistogramAggregation(ElasticsearchMetric):
    """
    Base elasticsearch histogram aggregation metric

    Attributes
    ----------
    field:
        The histogram field
    script:
        If provided, it will be used as scripted field
        for aggregation
    fixed_interval:
        If provided, it will used ALWAYS as the histogram
        aggregation interval
    """

    field: str
    script: Optional[Union[str, Dict[str, Any]]] = None
    fixed_interval: Optional[float] = None

    def aggregation_request(self, interval: float) -> Dict[str, Any]:
        if self.fixed_interval:
            interval = self.fixed_interval
        return {
            self.id: aggregations.histogram_aggregation(
                field_name=self.field, script=self.script, interval=interval
            )
        }


class TermsAggregation(ElasticsearchMetric):
    """
    The base elasticsearch terms aggregation metric

    Attributes
    ----------

    field:
        The term field
    script:
        If provided, it will be used as scripted field
        for aggregation
    fixed_size:
        If provided, the size will use for terms aggregation
    missing:
        If provided, will use the value for docs results with missing value for field

    """

    field: str = None
    script: Union[str, Dict[str, Any]] = None
    fixed_size: Optional[int] = None
    missing: Optional[str] = None

    def aggregation_request(self, size: int = None) -> Dict[str, Any]:
        if self.fixed_size:
            size = self.fixed_size
        return {
            self.id: aggregations.terms_aggregation(
                self.field, script=self.script, size=size, missing=self.missing
            )
        }



import rubrix as rb
from tests.monitoring.helpers import mock_monitor
from tests.server.test_helpers import client, mocking_client


def test_spacy_ner_monitor(monkeypatch):
    mocking_client(monkeypatch, client)
    dataset = "spacy-dataset"
    rb.delete(dataset)

    import spacy

    nlp = spacy.load("en_core_web_sm")
    nlp = rb.monitor(nlp, dataset=dataset, sample_rate=0.1)
    mock_monitor(nlp, monkeypatch)

    for _ in range(0, 100):
        nlp("Paris is my favourite city")

    df = rb.load(dataset)
    assert 1 < len(df) <= 20
    assert df.text.unique().tolist() == ["Paris is my favourite city"]

    rb.delete(dataset)
    list(nlp.pipe(["This is a text"] * 100))

    df = rb.load(dataset)
    assert 1 < len(df) <= 20
    assert df.text.unique().tolist() == ["This is a text"]

    rb.delete(dataset)
    list(nlp.pipe([("This is a text", {"meta": "data"})] * 100, as_tuples=True))

    df = rb.load(dataset)
    assert 1 < len(df) <= 20
    for metadata in df.metadata.values.tolist():
        assert metadata == {"meta": "data"}



from typing import List, Optional

from pydantic import BaseModel


class User(BaseModel):
    """User data model"""

    username: str
    email: Optional[str] = None
    full_name: Optional[str] = None
    disabled: Optional[bool] = None
    workspaces: List[str] = None




import pytest
from fastapi.security import SecurityScopes

from rubrix.server.security.auth_provider.local.provider import create_local_auth_provider

localAuth = create_local_auth_provider()
security_Scopes = SecurityScopes


# Tests for function get_user via token and api key
@pytest.mark.asyncio
async def test_get_user_via_token():
    access_token = localAuth._create_access_token(username="rubrix")
    user = await localAuth.get_user(security_scopes=security_Scopes, token=access_token)
    assert user.username == "rubrix"


@pytest.mark.asyncio
async def test_get_user_via_api_key():
    user = await localAuth.get_user(security_scopes=security_Scopes,api_key="rubrix.apikey")
    assert user.username == "rubrix"


# Test for function find user by api key token
@pytest.mark.asyncio
async def test_get_user_by_api_key():
    user = await localAuth._find_user_by_api_key(api_key="rubrix.apikey")
    assert user.username == "rubrix"


# Test for function fetch token
def test_fetch_token_user():
    access_token = localAuth._create_access_token(username="rubrix")
    user = localAuth.fetch_token_user(token=access_token)
    assert user.username == "rubrix"




from typing import Any, List, Optional, Type

from pydantic import BaseModel

from rubrix.server.commons.errors import WrongTaskError
from rubrix.server.tasks.commons import BaseRecord, TaskType
from rubrix.server.tasks.commons.metrics.model.base import BaseTaskMetrics


class TaskConfig(BaseModel):
    task: TaskType
    query: Any
    record: Type[BaseRecord]
    metrics: Optional[Type[BaseTaskMetrics]]


class TaskFactory:

    _REGISTERED_TASKS = dict()

    @classmethod
    def register_task(
        cls,
        task_type: TaskType,
        query_request: Type[Any],
        record_class: Type[BaseRecord],
        metrics: Optional[Type[BaseTaskMetrics]] = None,
    ):
        if metrics:
            metrics.configure_es_index()

        cls._REGISTERED_TASKS[task_type] = TaskConfig(
            task=task_type, query=query_request, record=record_class, metrics=metrics
        )

    @classmethod
    def get_all(cls) -> List[TaskConfig]:
        return [cfg for cfg in cls._REGISTERED_TASKS.values()]

    @classmethod
    def get_task_by_task_type(cls, task_type: TaskType) -> Optional[TaskConfig]:
        return cls._REGISTERED_TASKS.get(task_type)

    @classmethod
    def get_task_metrics(cls, task: TaskType) -> Optional[Type[BaseTaskMetrics]]:
        config = cls.get_task_by_task_type(task)
        if config:
            return config.metrics

    @classmethod
    def get_task_record(cls, task: TaskType) -> Type[BaseRecord]:
        config = cls.get_task_by_task_type(task)
        if not config:
            raise WrongTaskError(f"No configuration found for task {task}")
        return config.record



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import importlib
import sys

import numpy as np
import pytest

from rubrix import TextClassificationRecord

# we importlib.reload the `label_models` module during the tests, so avoid loading namespaces from this module!
# for example, don't do `from rubrix.labeling.text_classification.label_models import MissingAnnotationError`
# instead do `from rubrix.labeling.text_classification import label_models; label_models.MissingAnnotationError`
from rubrix.labeling.text_classification import Snorkel, WeakLabels, label_models


@pytest.fixture
def weak_labels(monkeypatch):
    def mock_load(*args, **kwargs):
        return [TextClassificationRecord(inputs="test", id=i) for i in range(4)]

    monkeypatch.setattr(
        "rubrix.labeling.text_classification.weak_labels.load", mock_load
    )

    def mock_apply(self, *args, **kwargs):
        weak_label_matrix = np.array(
            [[0, 1, -1], [2, 0, -1], [-1, -1, -1], [0, 2, 2]],
            dtype=np.short,
        )
        annotation_array = np.array([0, 1, -1, 2], dtype=np.short)
        label2int = {None: -1, "negative": 0, "positive": 1, "neutral": 2}
        return weak_label_matrix, annotation_array, label2int

    monkeypatch.setattr(WeakLabels, "_apply_rules", mock_apply)

    return WeakLabels(rules=[lambda: None] * 3, dataset="mock")


def test_weak_label_property():
    weak_labels = object()
    label_model = label_models.LabelModel(weak_labels)

    assert label_model.weak_labels is weak_labels


def test_abstract_methods():
    label_model = label_models.LabelModel(None)
    with pytest.raises(NotImplementedError):
        label_model.fit()
    with pytest.raises(NotImplementedError):
        label_model.score()
    with pytest.raises(NotImplementedError):
        label_model.predict()


@pytest.fixture
def uninstall_snorkel(monkeypatch):
    saved_module = sys.modules["snorkel"]
    sys.modules["snorkel"] = None
    importlib.reload(label_models)
    yield
    sys.modules["snorkel"] = saved_module
    importlib.reload(label_models)


def test_snorkel_not_installed(uninstall_snorkel):
    with pytest.raises(ModuleNotFoundError, match="pip install snorkel"):
        Snorkel(None)


def test_snorkel_init(weak_labels):
    from snorkel.labeling.model import LabelModel as SnorkelLabelModel

    label_model = Snorkel(weak_labels)

    assert label_model.weak_labels is weak_labels
    assert isinstance(label_model._model, SnorkelLabelModel)
    assert label_model._model.cardinality == 3


@pytest.mark.parametrize(
    "wrong_mapping",
    [
        {None: -10, "negative": 0, "positive": 1, "neutral": 2},
        {None: -1, "negative": 1, "positive": 3, "neutral": 4},
    ],
)
def test_snorkel_init_wrong_mapping(weak_labels, wrong_mapping):
    weak_labels.change_mapping(wrong_mapping)
    label_model = Snorkel(weak_labels)

    assert label_model.weak_labels.label2int == {
        None: -1,
        "negative": 0,
        "positive": 1,
        "neutral": 2,
    }


@pytest.mark.parametrize(
    "include_annotated_records",
    [True, False],
)
def test_snorkel_fit(monkeypatch, weak_labels, include_annotated_records):
    def mock_fit(self, L_train, *args, **kwargs):
        if include_annotated_records:
            assert (L_train == weak_labels.matrix()).all()
        else:
            assert (L_train == weak_labels.matrix(has_annotation=False)).all()
        assert kwargs == {"passed_on": None}

    monkeypatch.setattr(
        "rubrix.labeling.text_classification.label_models.SnorkelLabelModel.fit",
        mock_fit,
    )

    label_model = Snorkel(weak_labels)
    label_model.fit(include_annotated_records=include_annotated_records, passed_on=None)


def test_snorkel_fit_automatically_added_kwargs(weak_labels):
    label_model = Snorkel(weak_labels)
    with pytest.raises(ValueError, match="provided automatically"):
        label_model.fit(L_train=None)


@pytest.mark.parametrize(
    "policy,include_annotated_records,include_abstentions,expected",
    [
        ("abstain", True, False, (2, ["positive", "negative"], [0.8, 0.9])),
        (
            "abstain",
            True,
            True,
            (4, [None, None, "positive", "negative"], [None, None, 0.8, 0.9]),
        ),
        ("random", False, True, (1, ["positive"], [0.8])),
        (
            "random",
            True,
            True,
            (
                4,
                ["positive", "negative", "positive", "negative"],
                [0.4 + 0.0001, 1.0 / 3 + 0.0001, 0.8, 0.9],
            ),
        ),
    ],
)
def test_snorkel_predict(
    weak_labels,
    monkeypatch,
    policy,
    include_annotated_records,
    include_abstentions,
    expected,
):
    def mock_predict(self, L, return_probs, tie_break_policy, *args, **kwargs):
        assert tie_break_policy == policy
        assert return_probs is True
        if include_annotated_records:
            assert len(L) == 4
            preds = np.array([-1, -1, 1, 0])
            if policy == "random":
                preds = np.array([1, 0, 1, 0])
            return preds, np.array(
                [
                    [0.4, 0.4, 0.2],
                    [1.0 / 3, 1.0 / 3, 1.0 / 3],
                    [0.1, 0.8, 0.1],
                    [0.9, 0.05, 0.05],
                ]
            )
        else:
            assert len(L) == 1
            return np.array([1]), np.array([[0.1, 0.8, 0.1]])

    monkeypatch.setattr(
        "rubrix.labeling.text_classification.label_models.SnorkelLabelModel.predict",
        mock_predict,
    )

    label_model = Snorkel(weak_labels)

    records = label_model.predict(
        tie_break_policy=policy,
        include_annotated_records=include_annotated_records,
        include_abstentions=include_abstentions,
    )
    assert len(records) == expected[0]
    assert [
        rec.prediction[0][0] if rec.prediction else None for rec in records
    ] == expected[1]
    assert [
        rec.prediction[0][1] if rec.prediction else None for rec in records
    ] == expected[2]


@pytest.mark.parametrize("policy,expected", [("abstain", 0.5), ("random", 2.0 / 3)])
def test_snorkel_score(monkeypatch, weak_labels, policy, expected):
    def mock_predict(self, L, return_probs, tie_break_policy):
        assert (L == weak_labels.matrix(has_annotation=True)).all()
        assert return_probs is True
        assert tie_break_policy == policy
        if policy == "abstain":
            predictions = np.array([-1, 1, 0])
        elif policy == "random":
            predictions = np.array([0, 1, 0])
        else:
            raise ValueError("Untested policy!")

        probabilities = None  # accuracy does not need probabs ...

        return predictions, probabilities

    monkeypatch.setattr(
        "rubrix.labeling.text_classification.label_models.SnorkelLabelModel.predict",
        mock_predict,
    )

    label_model = Snorkel(weak_labels)
    assert label_model.score(tie_break_policy=policy)["accuracy"] == pytest.approx(
        expected
    )


def test_snorkel_score_without_annotations(weak_labels):
    weak_labels._annotation_array = np.array([], dtype=np.short)
    label_model = Snorkel(weak_labels)

    with pytest.raises(
        label_models.MissingAnnotationError, match="need annotated records"
    ):
        label_model.score()


@pytest.fixture
def weak_labels_from_guide(monkeypatch, resources):
    matrix_and_annotation = np.load(
        str(resources / "weak-supervision-guide-matrix.npy")
    )
    matrix, annotation = matrix_and_annotation[:, :-1], matrix_and_annotation[:, -1]

    def mock_load(*args, **kwargs):
        return [
            TextClassificationRecord(inputs="mock", id=i) for i in range(len(matrix))
        ]

    monkeypatch.setattr(
        "rubrix.labeling.text_classification.weak_labels.load", mock_load
    )

    def mock_apply(self, *args, **kwargs):
        return matrix, annotation, {None: -1, "SPAM": 0, "HAM": 1}

    monkeypatch.setattr(WeakLabels, "_apply_rules", mock_apply)

    return WeakLabels(rules=[lambda x: "mock"] * matrix.shape[1], dataset="mock")


def test_snorkel_integration(weak_labels_from_guide):
    label_model = Snorkel(weak_labels_from_guide)
    label_model.fit(seed=43)

    metrics = label_model.score()
    assert metrics["accuracy"] == pytest.approx(0.8947368421052632)

    records = label_model.predict()
    assert len(records) == 1177
    assert records[0].prediction == [
        ("SPAM", pytest.approx(0.5633776670811805)),
        ("HAM", pytest.approx(0.4366223329188196)),
    ]



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from datetime import datetime

import httpx
import pytest

from rubrix.client.models import (
    TokenClassificationRecord as ClientTokenClassificationRecord,
)
from rubrix.client.sdk.commons.models import BulkResponse
from rubrix.client.sdk.token_classification.api import bulk, data
from rubrix.client.sdk.token_classification.models import (
    CreationTokenClassificationRecord,
    TokenClassificationBulkData,
    TokenClassificationRecord,
)
from tests.server.test_helpers import client


@pytest.fixture
def bulk_data():
    records = [
        ClientTokenClassificationRecord(
            text="a raw text",
            tokens=["a", "raw", "text"],
            prediction=[("test", 2, 5, 0.9)],
            prediction_agent="agent",
            annotation=[("test", 2, 5)],
            annotation_agent="agent",
            id=i,
            metadata={"mymetadata": "str"},
            event_timestamp=datetime(2020, 1, 1),
            status="Validated",
        )
        for i in range(3)
    ]

    return TokenClassificationBulkData(
        records=[CreationTokenClassificationRecord.from_client(rec) for rec in records],
        tags={"Mytag": "tag"},
        metadata={"MyMetadata": 5},
    )


def test_bulk(sdk_client, bulk_data, monkeypatch):
    monkeypatch.setattr(httpx, "post", client.post)

    dataset_name = "test_dataset"
    client.delete(f"/api/datasets/{dataset_name}")
    response = bulk(sdk_client, name=dataset_name, json_body=bulk_data)

    assert response.status_code == 200
    assert isinstance(response.parsed, BulkResponse)


@pytest.mark.parametrize("limit,expected", [(None, 3), (2, 2)])
def test_data(limit, expected, sdk_client, bulk_data, monkeypatch):
    # TODO: Not sure how to test the streaming part of the response here
    monkeypatch.setattr(httpx, "stream", client.stream)

    dataset_name = "test_dataset"
    client.delete(f"/api/datasets/{dataset_name}")
    client.post(
        f"/api/datasets/{dataset_name}/TokenClassification:bulk",
        json=bulk_data.dict(by_alias=True),
    )

    response = data(sdk_client, name=dataset_name, limit=limit)
    assert isinstance(response.parsed[0], TokenClassificationRecord)
    assert len(response.parsed) == expected



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional, Union

from pydantic import BaseModel, Field, root_validator, validator

from rubrix._constants import MAX_KEYWORD_LENGTH
from rubrix.server.commons.es_helpers import filters
from rubrix.server.commons.helpers import flatten_dict
from rubrix.server.datasets.model import UpdateDatasetRequest
from rubrix.server.tasks.commons.api.model import (
    BaseAnnotation,
    BaseRecord,
    BaseSearchResults,
    BaseSearchResultsAggregations,
    PredictionStatus,
    ScoreRange,
    SortableField,
    TaskStatus,
    TaskType,
)


class ClassPrediction(BaseModel):
    """
    Single class prediction

    Attributes:
    -----------

    class_label: Union[str, int]
        the predicted class

    score: float
        the predicted class score. For human-supervised annotations,
        this probability should be 1.0
    """

    class_label: Union[str, int] = Field(alias="class")
    score: float = Field(default=1.0, ge=0.0, le=1.0)

    @validator("class_label")
    def check_label_length(cls, class_label):
        if isinstance(class_label, str):
            assert 1 <= len(class_label) <= MAX_KEYWORD_LENGTH, (
                f"Class name '{class_label}' exceeds max length of {MAX_KEYWORD_LENGTH}"
                if len(class_label) > MAX_KEYWORD_LENGTH
                else f"Class name must not be empty"
            )
        return class_label

    # See <https://pydantic-docs.helpmanual.io/usage/model_config>
    class Config:
        allow_population_by_field_name = True


class TextClassificationAnnotation(BaseAnnotation):
    """
    Annotation class for text classification tasks

    Attributes:
    -----------

    labels: List[LabelPrediction]
        list of annotated labels with score
    """

    labels: List[ClassPrediction]

    @validator("labels")
    def sort_labels(cls, labels: List[ClassPrediction]):
        """Sort provided labels by score"""
        return sorted(labels, key=lambda x: x.score, reverse=True)


class TokenAttributions(BaseModel):
    """
    The token attributions explaining predicted labels

    Attributes:
    -----------

    token: str
        The input token
    attributions: Dict[str, float]
        A dictionary containing label class-attribution pairs

    """

    token: str
    attributions: Dict[str, float] = Field(default_factory=dict)


class CreationTextClassificationRecord(BaseRecord[TextClassificationAnnotation]):
    """
    Text classification record

    Attributes:
    -----------

    inputs: Dict[str, Union[str, List[str]]]
        The input data text

    multi_label: bool
        Enable text classification with multiple predicted/annotated labels.
        Default=False

    explanation: Dict[str, List[TokenAttributions]]
        Token attribution list explaining predicted classes per token input.
        The dictionary key must be aligned with provided record text. Optional
    """

    inputs: Dict[str, Union[str, List[str]]]
    multi_label: bool = False
    explanation: Optional[Dict[str, List[TokenAttributions]]] = None

    _SCORE_DEVIATION_ERROR: ClassVar[float] = 0.001

    @root_validator
    def validate_record(cls, values):
        """fastapi validator method"""
        prediction = values.get("prediction", None)
        annotation = values.get("annotation", None)
        status = values.get("status")
        multi_label = values.get("multi_label", False)

        cls._check_score_integrity(prediction, multi_label)
        cls._check_annotation_integrity(annotation, multi_label, status)

        return values

    @classmethod
    def _check_annotation_integrity(
        cls,
        annotation: TextClassificationAnnotation,
        multi_label: bool,
        status: TaskStatus,
    ):
        if status == TaskStatus.validated and not multi_label:
            assert (
                annotation and len(annotation.labels) > 0
            ), "Annotation must include some label for validated records"

        if not multi_label and annotation:
            assert len(annotation.labels) == 1, "Single label record must include only one annotation label"


    @classmethod
    def _check_score_integrity(
        cls, prediction: TextClassificationAnnotation, multi_label: bool
    ):
        """
        Checks the score value integrity

        Parameters
        ----------
        prediction:
            The prediction annotation
        multi_label:
            If multi label

        """
        if prediction and not multi_label:
            assert sum([label.score for label in prediction.labels]) <= (
                1.0 + cls._SCORE_DEVIATION_ERROR
            ), f"Wrong score distributions: {prediction.labels}"

    @classmethod
    def task(cls) -> TaskType:
        """The task type"""
        return TaskType.text_classification

    @property
    def predicted(self) -> Optional[PredictionStatus]:
        if self.predicted_as and self.annotated_as:
            return (
                PredictionStatus.OK
                if set(self.predicted_as) == set(self.annotated_as)
                else PredictionStatus.KO
            )
        return None

    @property
    def words(self) -> str:
        sentences = []
        for v in self.inputs.values():
            if isinstance(v, list):
                sentences.extend(v)
            else:
                sentences.append(v)
        return "\n".join(sentences)

    @property
    def predicted_as(self) -> List[str]:
        return self._labels_from_annotation(
            self.prediction, multi_label=self.multi_label
        )

    @property
    def annotated_as(self) -> List[str]:
        return self._labels_from_annotation(
            self.annotation, multi_label=self.multi_label
        )

    @property
    def scores(self) -> List[float]:
        """Values of prediction scores"""
        if not self.prediction:
            return []
        return (
            [label.score for label in self.prediction.labels]
            if self.multi_label
            else [
                prediction_class.score
                for prediction_class in [
                    self._max_class_prediction(
                        self.prediction, multi_label=self.multi_label
                    )
                ]
                if prediction_class
            ]
        )

    @validator("inputs")
    def validate_inputs(cls, text: Dict[str, Any]):
        """Applies validation over input text"""
        assert len(text) > 0, "No inputs provided"

        for t in text.values():
            assert t is not None, "Cannot include None fields"

        return text

    @validator("inputs")
    def flatten_text(cls, text: Dict[str, Any]):
        """Normalizes input text to dict of strings"""
        flat_dict = flatten_dict(text)
        return flat_dict

    @classmethod
    def _labels_from_annotation(
        cls, annotation: TextClassificationAnnotation, multi_label: bool
    ) -> Union[List[str], List[int]]:
        """
        Extracts labels values from annotation

        Parameters
        ----------
        annotation:
            The annotation
        multi_label
            Enable/Disable multi label model

        Returns
        -------
            Label values for a given annotation

        """
        if not annotation:
            return []

        if multi_label:
            return [
                label.class_label for label in annotation.labels if label.score > 0.5
            ]

        class_prediction = cls._max_class_prediction(
            annotation, multi_label=multi_label
        )
        if class_prediction is None:
            return []

        return [class_prediction.class_label]

    @staticmethod
    def _max_class_prediction(
        p: TextClassificationAnnotation, multi_label: bool
    ) -> Optional[ClassPrediction]:
        """
        Gets the max class prediction for annotation

        Parameters
        ----------
        p:
            The annotation
        multi_label:
            Enable/Disable multi_label mode

        Returns
        -------

            The max class prediction in terms of prediction score if
            prediction has labels and no multi label is enabled. None, otherwise
        """
        if multi_label or p is None or not p.labels:
            return None
        return p.labels[0]

    class Config:
        allow_population_by_field_name = True


class TextClassificationRecord(CreationTextClassificationRecord):
    """
    The main text classification task record

    Attributes:
    -----------

    last_updated: datetime
        Last record update (read only)
    predicted: Optional[PredictionStatus]
        The record prediction status. Optional
    """

    last_updated: datetime = None
    _predicted: Optional[PredictionStatus] = Field(alias="predicted")


class TextClassificationBulkData(UpdateDatasetRequest):
    """
    API bulk data for text classification

    Attributes:
    -----------

    records: List[CreationTextClassificationRecord]
        The text classification record list

    """

    records: List[CreationTextClassificationRecord]

    @validator("records")
    def check_multi_label_integrity(cls, records: List[TextClassificationRecord]):
        """Checks all records in batch have same multi-label configuration"""
        if records:
            multi_label = records[0].multi_label
            for record in records[1:]:
                assert (
                    multi_label == record.multi_label
                ), "All records must be single/multi labelled"
        return records


class TextClassificationQuery(BaseModel):
    """
    API Filters for text classification

    Attributes:
    -----------
    ids: Optional[List[Union[str, int]]]
        Record ids list

    query_text: str
        Text query over inputs
    metadata: Optional[Dict[str, Union[str, List[str]]]]
        Text query over metadata fields. Default=None

    predicted_as: List[str]
        List of predicted terms
    annotated_as: List[str]
        List of annotated terms
    annotated_by: List[str]
        List of annotation agents
    predicted_by: List[str]
        List of predicted agents
    status: List[TaskStatus]
        List of task status
    predicted: Optional[PredictionStatus]
        The task prediction status

    """

    ids: Optional[List[Union[str, int]]]

    query_text: str = Field(default=None)
    metadata: Optional[Dict[str, Union[str, List[str]]]] = None

    predicted_as: List[str] = Field(default_factory=list)
    annotated_as: List[str] = Field(default_factory=list)
    annotated_by: List[str] = Field(default_factory=list)
    predicted_by: List[str] = Field(default_factory=list)
    score: Optional[ScoreRange] = Field(default=None)
    status: List[TaskStatus] = Field(default_factory=list)
    predicted: Optional[PredictionStatus] = Field(default=None, nullable=True)

    def as_elasticsearch(self) -> Dict[str, Any]:
        """Build an elasticsearch query part from search query"""

        if self.ids:
            return {"ids": {"values": self.ids}}

        all_filters = filters.metadata(self.metadata)
        query_filters = [
            query_filter
            for query_filter in [
                filters.predicted_as(self.predicted_as),
                filters.predicted_by(self.predicted_by),
                filters.annotated_as(self.annotated_as),
                filters.annotated_by(self.annotated_by),
                filters.status(self.status),
                filters.predicted(self.predicted),
                filters.score(self.score),
            ]
            if query_filter
        ]
        query_text = filters.text_query(self.query_text)
        all_filters.extend(query_filters)

        return {
            "bool": {
                "must": query_text or {"match_all": {}},
                "filter": {
                    "bool": {
                        "should": all_filters,
                        "minimum_should_match": len(all_filters),
                    }
                },
            }
        }


class TextClassificationSearchRequest(BaseModel):
    """
    API SearchRequest request

    Attributes:
    -----------

    query: TextClassificationQuery
        The search query configuration

    sort:
        The sort order list
    """

    query: TextClassificationQuery = Field(default_factory=TextClassificationQuery)
    sort: List[SortableField] = Field(default_factory=list)


class TextClassificationSearchAggregations(BaseSearchResultsAggregations):
    pass


class TextClassificationSearchResults(
    BaseSearchResults[TextClassificationRecord, TextClassificationSearchAggregations]
):
    pass



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

from fastapi import APIRouter, Depends, Query, Security

from rubrix.server.security import auth
from rubrix.server.security.model import User
from .model import CopyDatasetRequest, Dataset, UpdateDatasetRequest
from .service import DatasetsService
from ..commons.api import CommonTaskQueryParams

router = APIRouter(tags=["datasets"], prefix="/datasets")


@router.get(
    "/",
    response_model=List[Dataset],
    response_model_exclude_none=True,
    operation_id="list_datasets",
)
def list_datasets(
    ds_params: CommonTaskQueryParams = Depends(),
    service: DatasetsService = Depends(DatasetsService.get_instance),
    current_user: User = Security(auth.get_user, scopes=[]),
) -> List[Dataset]:
    """
    List accessible user datasets

    Parameters
    ----------
    ds_params:
        Common task query params
    service:
        The datasets service
    current_user:
        The request user

    Returns
    -------
        A list of datasets visibles by current user
    """
    return service.list(
        user=current_user,
        workspaces=[ds_params.workspace],
    )


@router.get(
    "/{name}",
    response_model=Dataset,
    response_model_exclude_none=True,
    operation_id="get_dataset",
)
def get_dataset(
    name: str,
    ds_params: CommonTaskQueryParams = Depends(),
    service: DatasetsService = Depends(DatasetsService.get_instance),
    current_user: User = Security(auth.get_user, scopes=[]),
) -> Dataset:
    """
    Find a dataset by name visible for current user

    Parameters
    ----------
    name:
        The dataset name
    ds_params:
        Common dataset query params
    service:
        Datasets service
    current_user:
        The current user

    Returns
    -------
        - The found dataset if accessible or exists.
        - EntityNotFoundError if not found.
        - NotAuthorizedError if user cannot access the found dataset

    """
    return Dataset.parse_obj(
        service.find_by_name(
            name, task=None, user=current_user, workspace=ds_params.workspace
        )
    )


@router.patch(
    "/{name}",
    operation_id="update_dataset",
    response_model=Dataset,
    response_model_exclude_none=True,
)
def update_dataset(
    name: str,
    update_request: UpdateDatasetRequest,
    ds_params: CommonTaskQueryParams = Depends(),
    service: DatasetsService = Depends(DatasetsService.get_instance),
    current_user: User = Security(auth.get_user, scopes=[]),
) -> Dataset:
    """
    Update a set of parameters for a dataset

    Parameters
    ----------
    name:
        The dataset name
    update_request:
        The fields to update
    ds_params:
        Common dataset query params
    service:
        The datasets service
    current_user:
        The current user

    Returns
    -------

    - The updated dataset if exists and user has access.
    - EntityNotFoundError if not found.
    - NotAuthorizedError if user cannot access the found dataset

    """
    return service.update(
        name,
        data=update_request,
        user=current_user,
        workspace=ds_params.workspace,
    )


@router.delete(
    "/{name}",
    operation_id="delete_dataset",
)
def delete_dataset(
    name: str,
    ds_params: CommonTaskQueryParams = Depends(),
    service: DatasetsService = Depends(DatasetsService.get_instance),
    current_user: User = Security(auth.get_user, scopes=[]),
):
    """
    Deletes a dataset

    Parameters
    ----------
    name:
        The dataset name
    ds_params:
        Common dataset query params
    service:
        The datasets service
    current_user:
        The current user

    """
    service.delete(name, user=current_user, workspace=ds_params.workspace)


@router.put(
    "/{name}:close",
    operation_id="close_dataset",
)
def close_dataset(
    name: str,
    ds_params: CommonTaskQueryParams = Depends(),
    service: DatasetsService = Depends(DatasetsService.get_instance),
    current_user: User = Security(auth.get_user, scopes=[]),
):
    """
    Closes a dataset. This operation will releases backend resources

    Parameters
    ----------
    name:
        The dataset name
    ds_params:
        Common dataset query params
    service:
        The datasets service
    current_user:
        The current user

    """
    service.close_dataset(name, user=current_user, workspace=ds_params.workspace)


@router.put(
    "/{name}:open",
    operation_id="open_dataset",
)
def open_dataset(
    name: str,
    ds_params: CommonTaskQueryParams = Depends(),
    service: DatasetsService = Depends(DatasetsService.get_instance),
    current_user: User = Security(auth.get_user, scopes=[]),
):
    """
    Closes a dataset. This operation will releases backend resources

    Parameters
    ----------
    name:
        The dataset name
    ds_params:
        Common dataset query params
    service:
        The datasets service
    current_user:
        The current user

    """
    service.open_dataset(name, user=current_user, workspace=ds_params.workspace)


@router.put(
    "/{name}:copy",
    operation_id="copy_dataset",
    response_model=Dataset,
    response_model_exclude_none=True,
)
def copy_dataset(
    name: str,
    copy_request: CopyDatasetRequest,
    ds_params: CommonTaskQueryParams = Depends(),
    service: DatasetsService = Depends(DatasetsService.get_instance),
    current_user: User = Security(auth.get_user, scopes=[]),
) -> Dataset:
    """
    Creates a dataset copy and its tags/metadata info

    Parameters
    ----------
    name:
        The dataset name
    copy_request:
        The copy request data
    ds_params:
        Common dataset query params
    service:
        The datasets service
    current_user:
        The current user

    """

    return service.copy_dataset(
        name=name, data=copy_request, user=current_user, workspace=ds_params.workspace
    )



from rubrix.server.tasks.text2text import (
    Text2TextAnnotation,
    Text2TextPrediction,
    Text2TextQuery,
    Text2TextRecord,
)


def test_sentences_sorted_by_score():

    record = Text2TextRecord(
        text="The inpu2 text",
        prediction=Text2TextAnnotation(
            agent="test_sentences_sorted_by_score",
            sentences=[
                Text2TextPrediction(text="sentence 1", score=0.6),
                Text2TextPrediction(text="sentence 2", score=0.5),
                Text2TextPrediction(text="sentence 3", score=1.0),
            ],
        ),
    )

    assert record.prediction.sentences[0].score == 1.0
    assert record.prediction.sentences[1].score == 0.6
    assert record.prediction.sentences[2].score == 0.5


def test_query_as_elasticsearch():
    query = Text2TextQuery(ids=[1, 2, 3])
    assert query.as_elasticsearch() == {"ids": {"values": query.ids}}



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest

from rubrix.server.tasks.commons import BulkResponse, SortableField
from rubrix.server.tasks.token_classification import (
    TokenClassificationQuery,
    TokenClassificationSearchRequest,
    TokenClassificationSearchResults,
)
from rubrix.server.tasks.token_classification.api import (
    TokenClassificationBulkData,
    TokenClassificationRecord,
)
from tests.server.test_helpers import client


def test_load_as_different_task():
    dataset = "test_load_as_different_task"
    assert client.delete(f"/api/datasets/{dataset}").status_code == 200
    expected_text = "This is a text with !"
    records = [
        TokenClassificationRecord.parse_obj(data)
        for data in [
            {
                "tokens": expected_text.split(" "),
                "raw_text": expected_text,
            }
        ]
    ]
    client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkData(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
    )

    response = client.post(
        f"/api/datasets/{dataset}/TextClassification:search",
        json={},
    )
    assert response.status_code == 400


def test_search_special_characters():
    dataset = "test_search_special_characters"
    assert client.delete(f"/api/datasets/{dataset}").status_code == 200
    expected_text = "This is a text with !"
    records = [
        TokenClassificationRecord.parse_obj(data)
        for data in [
            {
                "tokens": expected_text.split(" "),
                "raw_text": expected_text,
            }
        ]
    ]
    client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkData(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
    )

    response = client.post(
        f"/api/datasets/{dataset}/TokenClassification:search",
        json=TokenClassificationSearchRequest(
            query=TokenClassificationQuery(query_text="\!")
        ).dict(),
    )
    assert response.status_code == 200, response.json()
    results = TokenClassificationSearchResults.parse_obj(response.json())
    assert results.total == 1


def test_some_sort():
    dataset = "test_some_sort"
    assert client.delete(f"/api/datasets/{dataset}").status_code == 200
    expected_text = "This is a text with !"
    records = [
        TokenClassificationRecord.parse_obj(data)
        for data in [
            {
                "tokens": expected_text.split(" "),
                "raw_text": expected_text,
            }
        ]
    ]
    client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkData(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
    )

    with pytest.raises(AssertionError):
        client.post(
            f"/api/datasets/{dataset}/TokenClassification:search",
            json=TokenClassificationSearchRequest(
                sort=[SortableField(id="babba")],
            ).dict(),
        )


def test_create_records_for_token_classification():
    dataset = "test_create_records_for_token_classification"
    assert client.delete(f"/api/datasets/{dataset}").status_code == 200
    entity_label = "TEST"

    records = [
        TokenClassificationRecord.parse_obj(data)
        for data in [
            {
                "tokens": "This is a text".split(" "),
                "raw_text": "This is a text",
                "metadata": {"field_one": "value one", "field_two": "value 2"},
                "prediction": {
                    "agent": "test",
                    "entities": [{"start": 0, "end": 4, "label": entity_label}],
                },
            },
            {
                "tokens": "This is a text".split(" "),
                "raw_text": "This is a text",
                "metadata": {"field_one": "value one", "field_two": "value 2"},
                "annotation": {
                    "agent": "test",
                    "entities": [{"start": 0, "end": 4, "label": entity_label}],
                },
            },
        ]
    ]
    response = client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkData(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
    )

    assert response.status_code == 200, response.json()
    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.dataset == dataset
    assert bulk_response.failed == 0
    assert bulk_response.processed == 2

    response = client.post(f"/api/datasets/{dataset}/TokenClassification:search")
    assert response.status_code == 200, response.json()
    results = TokenClassificationSearchResults.parse_obj(response.json())
    assert "This" in results.aggregations.predicted_mentions[entity_label]
    assert "This" in results.aggregations.mentions[entity_label]


def test_multiple_mentions_in_same_record():
    dataset = "test_multiple_mentions_in_same_record"
    assert client.delete(f"/api/datasets/{dataset}").status_code == 200
    entity_a, entity_b = ("TESTA", "TESTB")

    records = [
        TokenClassificationRecord.parse_obj(data)
        for data in [
            {
                "tokens": "This is a text".split(" "),
                "raw_text": "This is a text",
                "metadata": {"field_one": "value one", "field_two": "value 2"},
                "prediction": {
                    "agent": "test",
                    "entities": [
                        {"start": 0, "end": 4, "label": entity_a},
                        {"start": 5, "end": 7, "label": entity_b},
                    ],
                },
            },
            {
                "tokens": "This is a text".split(" "),
                "raw_text": "This is a text",
                "metadata": {"field_one": "value one", "field_two": "value 2"},
                "prediction": {
                    "agent": "test",
                    "entities": [{"start": 0, "end": 4, "label": entity_a}],
                },
            },
            {
                "tokens": "This is a text".split(" "),
                "raw_text": "This is a text",
                "metadata": {"field_one": "value one", "field_two": "value 2"},
                "annotation": {
                    "agent": "test",
                    "entities": [{"start": 0, "end": 4, "label": entity_a}],
                },
            },
        ]
    ]
    response = client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkData(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
    )

    assert response.status_code == 200, response.json()
    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.dataset == dataset
    assert bulk_response.failed == 0
    assert bulk_response.processed == 3

    response = client.post(f"/api/datasets/{dataset}/TokenClassification:search")
    assert response.status_code == 200, response.json()
    results = TokenClassificationSearchResults.parse_obj(response.json())
    assert "This" in results.aggregations.predicted_mentions[entity_a]
    assert results.aggregations.predicted_mentions[entity_a]["This"] == 2
    assert "is" in results.aggregations.predicted_mentions[entity_b]
    assert results.aggregations.predicted_mentions[entity_b]["is"] == 1


def test_show_not_aggregable_metadata_fields():
    dataset = "test_show_not_aggregable_metadata_fields"
    assert client.delete(f"/api/datasets/{dataset}").status_code == 200

    response = client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkData(
            records=[
                TokenClassificationRecord.parse_obj(
                    {
                        "tokens": "This is a text".split(" "),
                        "raw_text": "This is a text",
                        "metadata": {"field_one": 1.3, "field_two": 2},
                    },
                ),
                TokenClassificationRecord.parse_obj(
                    {
                        "tokens": "This is a text".split(" "),
                        "raw_text": "This is a text",
                        "metadata": {"field_one": 2.3, "field_two": 300},
                    },
                ),
                TokenClassificationRecord.parse_obj(
                    {
                        "tokens": "This is a text".split(" "),
                        "raw_text": "This is a text",
                        "metadata": {"field_one": 11.333, "field_two": -10},
                    },
                ),
            ],
        ).dict(by_alias=True),
    )

    assert response.status_code == 200, response.json()

    response = client.post(
        f"/api/datasets/{dataset}/TokenClassification:search",
        json={},
    )
    assert response.status_code == 200, response.json()
    results = TokenClassificationSearchResults.parse_obj(response.json())
    assert len(results.aggregations.metadata) == 2
    assert "field_one" in results.aggregations.metadata
    assert "rubrix:stats" in results.aggregations.metadata["field_one"]



DATASETS_INDEX_NAME = f".rubrix.datasets-v0"
DATASETS_RECORDS_INDEX_NAME = ".rubrix.dataset.{}.records-v0"


DATASETS_INDEX_TEMPLATE = {
    "index_patterns": [DATASETS_INDEX_NAME],
    "settings": {"number_of_shards": 1},
    "mappings": {
        "properties": {
            "tags": {
                "type": "nested",
                "properties": {
                    "key": {"type": "keyword"},
                    "value": {"type": "text"},
                },
            },
            "metadata": {
                "type": "nested",
                "properties": {
                    "key": {"type": "keyword"},
                    "value": {"type": "text"},
                },
            },
        }
    },
}




#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from enum import Enum
from typing import Any, Optional
from typing import Dict

from pydantic import BaseModel
from pydantic import Field


class TaskType(str, Enum):
    text_classification = "TextClassification"
    token_classification = "TokenClassification"
    text2text = "Text2Text"
    multi_task_text_token_classification = "MultitaskTextTokenClassification"


class BaseDatasetModel(BaseModel):
    tags: Dict[str, str] = Field(default_factory=dict)
    metadata: Dict[str, Any] = Field(default_factory=dict)
    name: str = Field(regex="^(?!-|_)[a-z0-9-_]+$")


class Dataset(BaseDatasetModel):
    task: TaskType
    owner: str = None
    created_at: datetime = None
    last_updated: datetime = None


class CopyDatasetRequest(BaseDatasetModel):
    target_workspace: Optional[str] = None



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Iterable, List, Optional

from fastapi import Depends

from rubrix.server.commons.es_helpers import sort_by2elasticsearch
from rubrix.server.datasets.model import Dataset
from rubrix.server.tasks.commons import (
    BulkResponse,
    EsRecordDataFieldNames,
    SortableField,
)
from rubrix.server.tasks.commons.dao import extends_index_dynamic_templates
from rubrix.server.tasks.commons.dao.dao import DatasetRecordsDAO, dataset_records_dao
from rubrix.server.tasks.commons.dao.model import RecordSearch
from rubrix.server.tasks.commons.metrics.service import MetricsService
from rubrix.server.tasks.text_classification.api.model import (
    CreationTextClassificationRecord,
    TextClassificationQuery,
    TextClassificationRecord,
    TextClassificationSearchAggregations,
    TextClassificationSearchResults,
)

extends_index_dynamic_templates(
    {"inputs": {"path_match": "inputs.*", "mapping": {"type": "text"}}}
)


class TextClassificationService:
    """
    Text classification service

    """

    def __init__(
        self,
        dao: DatasetRecordsDAO,
        metrics: MetricsService,
    ):
        self.__dao__ = dao
        self.__metrics__ = metrics

    def add_records(
        self,
        dataset: Dataset,
        records: List[CreationTextClassificationRecord],
    ):
        self._check_multi_label_integrity(dataset, records)
        self.__metrics__.build_records_metrics(dataset, records)
        failed = self.__dao__.add_records(
            dataset=dataset, records=records, record_class=TextClassificationRecord
        )
        return BulkResponse(dataset=dataset.name, processed=len(records), failed=failed)

    def search(
        self,
        dataset: Dataset,
        query: TextClassificationQuery,
        sort_by: List[SortableField],
        record_from: int = 0,
        size: int = 100,
    ) -> TextClassificationSearchResults:
        """
        Run a search in a dataset

        Parameters
        ----------
        dataset:
            The records dataset
        query:
            The search parameters
        sort_by:
            The sort by list
        record_from:
            The record from return results
        size:
            The max number of records to return

        Returns
        -------
            The matched records with aggregation info for specified task_meta.py

        """
        results = self.__dao__.search_records(
            dataset,
            search=RecordSearch(
                query=query.as_elasticsearch(),
                sort=sort_by2elasticsearch(
                    sort_by,
                    valid_fields=[
                        "metadata",
                        EsRecordDataFieldNames.score,
                        EsRecordDataFieldNames.predicted,
                        EsRecordDataFieldNames.predicted_as,
                        EsRecordDataFieldNames.predicted_by,
                        EsRecordDataFieldNames.annotated_as,
                        EsRecordDataFieldNames.annotated_by,
                        EsRecordDataFieldNames.status,
                        EsRecordDataFieldNames.event_timestamp,
                    ],
                ),
            ),
            size=size,
            record_from=record_from,
        )
        return TextClassificationSearchResults(
            total=results.total,
            records=[TextClassificationRecord.parse_obj(r) for r in results.records],
            aggregations=TextClassificationSearchAggregations(
                **results.aggregations,
                words=results.words,
                metadata=results.metadata or {},
            )
            if results.aggregations
            else None,
        )

    def read_dataset(
        self,
        dataset: Dataset,
        query: Optional[TextClassificationQuery] = None,
    ) -> Iterable[TextClassificationRecord]:
        """
        Scan a dataset records

        Parameters
        ----------
        dataset:
            The dataset name
        query:
            If provided, scan will retrieve only records matching
            the provided query filters. Optional

        """
        for db_record in self.__dao__.scan_dataset(
            dataset, search=RecordSearch(query=query.as_elasticsearch())
        ):
            yield TextClassificationRecord.parse_obj(db_record)

    def _check_multi_label_integrity(
        self, dataset: Dataset, records: List[TextClassificationRecord]
    ):
        # Fetch a single record
        results = self.search(
            dataset, query=TextClassificationQuery(), size=1, sort_by=[]
        )
        if results.records:
            is_multi_label = records[0].multi_label
            assert is_multi_label == results.records[0].multi_label, (
                "You cannot pass {labels_type} records for this dataset. "
                "Stored records are {labels_type}".format(
                    labels_type="multi-label" if is_multi_label else "single-label"
                )
            )


_instance = None


def text_classification_service(
    dao: DatasetRecordsDAO = Depends(dataset_records_dao),
    metrics: MetricsService = Depends(MetricsService.get_instance),
) -> TextClassificationService:
    """
    Creates a dataset record service instance

    Parameters
    ----------
    dao:
        The dataset records dao dependency
    metrics:
        The metrics service

    Returns
    -------
        A dataset records service instance
    """
    global _instance
    if not _instance:
        _instance = TextClassificationService(dao=dao, metrics=metrics)
    return _instance



from datetime import datetime
from typing import Any, Dict, Optional

import rubrix as rb
from rubrix import TokenClassificationRecord
from rubrix.monitoring.base import BaseMonitor
from rubrix.monitoring.types import MissingType

try:
    # Conditional modules
    from spacy import Language
    from spacy.tokens import Doc
except ModuleNotFoundError:
    Language = MissingType
    Doc = MissingType


class SpacyNERMonitor(BaseMonitor):
    """A spaCy Language wrapper for NLP NER monitoring in Rubrix"""

    @staticmethod
    def doc2token_classification(
        doc: Doc, agent: str, metadata: Optional[Dict[str, Any]]
    ) -> TokenClassificationRecord:
        """
        Converts a spaCy `Doc` into a token classification record

        Parameters
        ----------
        doc:
            The spacy doc
        agent:
            Agent to use for the prediction_agent field. Could be the model path or model lang + model version
        metadata:
            Passed on to the `rubrix.TokenClassificationRecord`.

        """
        entities = [(ent.label_, ent.start_char, ent.end_char) for ent in doc.ents]
        return TokenClassificationRecord(
            text=doc.text,
            tokens=[t.text for t in doc],
            metadata=metadata or {},
            prediction_agent=agent,
            prediction=entities,
            event_timestamp=datetime.utcnow(),
        )

    def _log2rubrix(self, doc: Doc, metadata: Optional[Dict[str, Any]] = None):
        record = self.doc2token_classification(
            doc, agent=self.__wrapped__.path.name, metadata=metadata
        )
        rb.log(
            record,
            name=self.dataset,
            tags={k: v for k, v in self.__wrapped__.meta.items() if isinstance(v, str)},
            metadata=self.__model__.meta,
            verbose=False,
        )

    def pipe(self, *args, **kwargs):
        as_tuples = kwargs.get("as_tuples")
        results = self.__model__.pipe(*args, **kwargs)

        for r in results:
            metadata = {}
            if as_tuples:
                doc, metadata = r  # context
            else:
                doc = r
            if self.is_record_accepted():
                self.log_async(doc, metadata)
            yield r

    def __call__(self, *args, **kwargs):
        metadata = kwargs.pop("metadata", None)
        doc = self.__wrapped__(*args, **kwargs)
        try:
            if self.is_record_accepted():
                self.log_async(doc, metadata)
        finally:
            return doc


def ner_monitor(nlp: Language, dataset: str, sample_rate: float) -> Language:
    return SpacyNERMonitor(nlp, dataset=dataset, sample_rate=sample_rate)



import warnings
from typing import Any, Callable, Dict

from pydantic import BaseModel, PrivateAttr


class MetricSummary(BaseModel):
    """THe metric summary result data model"""

    data: Dict[str, Any]
    _build_visualization: Callable = PrivateAttr()

    def visualize(self):
        try:
            return self._build_visualization()
        except ModuleNotFoundError:
            warnings.warn(
                "Please, install plotly in order to use this feature\n"
                "%>pip install plotly"
            )

    @classmethod
    def new_summary(
        cls, data: Dict[str, Any], visualization: Callable
    ) -> "MetricSummary":
        summary = cls(data=data)
        summary._build_visualization = visualization
        return summary



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest
from httpx import Response as HttpxResponse

from rubrix.client.sdk.commons.api import build_bulk_response, build_data_response
from rubrix.client.sdk.commons.models import (
    BulkResponse,
    ErrorMessage,
    HTTPValidationError,
    Response,
    ValidationError,
)
from rubrix.client.sdk.text_classification.models import TextClassificationRecord


@pytest.mark.parametrize(
    "status_code, expected",
    [
        (200, BulkResponse),
        (404, ErrorMessage),
        (500, ErrorMessage),
        (422, HTTPValidationError),
    ],
)
def test_build_bulk_response(status_code, expected):
    server_response = None
    if status_code == 200:
        server_response = BulkResponse(dataset="test", failed=0, processed=0)
    elif status_code == 404:
        server_response = ErrorMessage(detail="test")
    elif status_code == 500:
        server_response = ErrorMessage(detail="test")
    elif status_code == 422:
        server_response = HTTPValidationError(
            detail=[ValidationError(loc=["test"], msg="test", type="test")]
        )

    httpx_response = HttpxResponse(
        status_code=status_code, content=server_response.json()
    )
    response = build_bulk_response(httpx_response)

    assert isinstance(response, Response)
    assert isinstance(response.parsed, expected)


@pytest.mark.parametrize(
    "status_code, expected",
    [
        (200, TextClassificationRecord),
        (404, ErrorMessage),
        (500, ErrorMessage),
        (422, HTTPValidationError),
    ],
)
def test_build_data_response(status_code, expected):
    server_response = None
    if status_code == 200:
        server_response = TextClassificationRecord(inputs={"text": "test"})
    elif status_code == 404:
        server_response = ErrorMessage(detail="test")
    elif status_code == 500:
        server_response = ErrorMessage(detail="test")
    elif status_code == 422:
        server_response = HTTPValidationError(
            detail=[ValidationError(loc=["test"], msg="test", type="test")]
        )

    httpx_response = HttpxResponse(
        status_code=status_code,
        content=server_response.json() + "\n" + server_response.json(),
    )
    response = build_data_response(httpx_response, data_type=TextClassificationRecord)

    assert isinstance(response, Response)
    if status_code == 200:
        assert isinstance(response.parsed, list) and len(response.parsed) == 2
        assert isinstance(response.parsed[0], expected)
    else:
        assert isinstance(response.parsed, expected)



from dataclasses import dataclass

from fastapi import Header, Query

from rubrix._constants import RUBRIX_WORKSPACE_HEADER_NAME


@dataclass
class CommonTaskQueryParams:
    """Common task query params"""

    __workspace_param__: str = Query(
        None,
        alias="workspace",
        description="The workspace where dataset belongs to. If not provided default user team will be used",
    )

    __workspace_header__: str = Header(None, alias=RUBRIX_WORKSPACE_HEADER_NAME)

    @property
    def workspace(self) -> str:
        """Return read workspace. Query param prior to header param"""
        return self.__workspace_param__ or self.__workspace_header__



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import httpx
import pytest

from rubrix import load
from rubrix.client.models import TextClassificationRecord
from rubrix.client.sdk.text_classification.models import (
    CreationTextClassificationRecord,
    TextClassificationBulkData,
)
from rubrix.labeling.text_classification.rule import Rule, RuleNotAppliedError
from tests.server.test_helpers import client


@pytest.fixture(scope="module")
def log_dataset() -> str:
    dataset_name = "test_dataset_for_rule"
    client.delete(f"/api/datasets/{dataset_name}")
    records = [
        CreationTextClassificationRecord.parse_obj(
            {
                "inputs": {"text": text},
                "annotation": {
                    "labels": [{"class": label, "score": 1}],
                    "agent": "test",
                },
                "id": idx,
            }
        )
        for text, label, idx in zip(
            ["negative", "positive"], ["negative", "positive"], [1, 2]
        )
    ]
    client.post(
        f"/api/datasets/{dataset_name}/TextClassification:bulk",
        json=TextClassificationBulkData(
            records=records,
        ).dict(by_alias=True),
    )

    return dataset_name


@pytest.mark.parametrize(
    "name,expected", [(None, "query_string"), ("test_name", "test_name")]
)
def test_name(name, expected):
    rule = Rule(query="query_string", label="mock", name=name)
    assert rule.name == expected


def test_apply(monkeypatch, log_dataset):
    rule = Rule(query="inputs.text:(NOT positive)", label="negative")
    with pytest.raises(RuleNotAppliedError):
        rule(TextClassificationRecord(inputs="test"))

    monkeypatch.setattr(httpx, "get", client.get)
    monkeypatch.setattr(httpx, "stream", client.stream)

    rule.apply(log_dataset)
    assert rule._matching_ids == {1: None}


def test_call(monkeypatch, log_dataset):
    monkeypatch.setattr(httpx, "get", client.get)
    monkeypatch.setattr(httpx, "stream", client.stream)

    rule = Rule(query="inputs.text:(NOT positive)", label="negative")
    rule.apply(log_dataset)

    records = load(log_dataset, as_pandas=False)
    assert rule(records[0]) == "negative"
    assert rule(records[1]) is None



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Optional

import rubrix as rb
from rubrix import TextClassificationRecord


class Rule:
    """A rule (labeling function) in form of an ElasticSearch query.

    Args:
        query: An ElasticSearch query with the `query string syntax <https://rubrix.readthedocs.io/en/stable/reference/rubrix_webapp_reference.html#search-input>`_.
        label: The label associated to the query.
        name: An optional name for the rule to be used as identifier in the
            `rubrix.labeling.text_classification.WeakLabels` class. By default, we will use the ``query`` string.

    Examples:
        >>> import rubrix as rb
        >>> urgent_rule = Rule(query="inputs.text:(urgent AND immediately)", label="urgent", name="urgent_rule")
        >>> not_urgent_rule = Rule(query="inputs.text:(NOT urgent) AND metadata.title_length>20", label="not urgent")
        >>> not_urgent_rule.apply("my_dataset")
        >>> my_dataset_records = rb.load(name="my_dataset", as_pandas=False)
        >>> not_urgent_rule(my_dataset_records[0])
        "not urgent"
    """

    def __init__(self, query: str, label: str, name: Optional[str] = None):
        self._query = query
        self._label = label
        self._name = name
        self._matching_ids = None

    @property
    def name(self):
        """The name of the rule."""
        if self._name is not None:
            return self._name
        return self._query

    def apply(self, dataset: str):
        """Apply the rule to a dataset and save matching ids of the records.

        Args:
            dataset: The name of the dataset.
        """
        records = rb.load(name=dataset, query=self._query, as_pandas=False)

        self._matching_ids = {record.id: None for record in records}

    def __call__(self, record: TextClassificationRecord) -> Optional[str]:
        """Check if the given record is among the matching ids from the ``self.apply`` call.

        Args:
            record: The record to be labelled.

        Returns:
            A label if the record id is among the matching ids, otherwise None.

        Raises:
            RuleNotAppliedError: If the rule was not applied to the dataset before.
        """
        if self._matching_ids is None:
            raise RuleNotAppliedError(
                "Rule was still not applied. Please call `self.apply(dataset)` first."
            )

        try:
            self._matching_ids[record.id]
        except KeyError:
            return None
        else:
            return self._label


class RuleNotAppliedError(Exception):
    pass



import pytest

import rubrix
from rubrix import TokenClassificationRecord
from tests.server.test_helpers import client, mocking_client


def test_log_with_empty_text(monkeypatch):
    mocking_client(monkeypatch, client)
    dataset = "test_log_with_empty_text"
    text = " "

    rubrix.delete(dataset)
    with pytest.raises(Exception, match="No text or empty text provided"):
        rubrix.log(
            TokenClassificationRecord(id=0, text=text, tokens=["a", "b", "c"]),
            name=dataset,
        )


def test_log_with_empty_tokens_list(monkeypatch):
    mocking_client(monkeypatch, client)
    dataset = "test_log_with_empty_text"
    text = "The text"

    rubrix.delete(dataset)
    with pytest.raises(
        Exception,
        match="ensure this value has at least 1 items",
    ):
        rubrix.log(
            TokenClassificationRecord(id=0, text=text, tokens=[]),
            name=dataset,
        )


def test_log_record_that_makes_me_cry(monkeypatch):
    mocking_client(monkeypatch, client)
    dataset = "test_log_record_that_makes_me_cry"
    record = TokenClassificationRecord(
        text="'Secret Story : Última hora' debuta con un pobre 8.7% en el access de Telecinco.. . PROGRAMAS CON MEJOR CUOTA DEL LUNES (POR CADENAS). . ",
        tokens=[
            "'",
            "Secret",
            "Story",
            ":",
            "Última",
            "hora",
            "'",
            "debuta",
            "con",
            "un",
            "pobre",
            "8.7%",
            "en",
            "el",
            "access",
            "de",
            "Telecinco",
            "..",
            ".",
            "PROGRAMAS",
            "CON",
            "MEJOR",
            "CUOTA",
            "DEL",
            "LUNES",
            "(",
            "POR",
            "CADENAS",
            ")",
            ".",
            ".",
        ],
        prediction=[("ENG", 60, 66)],
        annotation=None,
        prediction_agent=None,
        annotation_agent=None,
        id=None,
        metadata={"section": "television", "newspaper": "eldiario"},
        status="Default",
        event_timestamp=None,
    )
    rubrix.delete(dataset)
    rubrix.log(record, name=dataset)

    records = rubrix.load(dataset, as_pandas=False)
    assert len(records) == 1
    assert records[0].text == record.text
    assert records[0].tokens == record.tokens
    assert records[0].metrics == {
        "annotated": {"mentions": []},
        "predicted": {
            "mentions": [
                {
                    "capitalness": "LOWER",
                    "chars_length": 6,
                    "density": 0.03225806451612903,
                    "label": "ENG",
                    "score": 1.0,
                    "tokens_length": 1,
                    "value": "access",
                }
            ]
        },
        "tokens_length": 31,
    }
